/*
 *  (C) Robert Bosch GmbH 2008. All rights reserved, also regarding any
 *  disposal, exploitation, reproduction, editing, distribution, as
 *  well as in the event of application for industrial property rights.
 *
 *  Contact: AE/NE1-IFS <vAENE1.IFS@de.bosch.com>
 *
 * Disclaimer:
 * This Software is licensed under the following conditions. You may 
 * not use this file except in compliance with the following conditions:
 *
 * Robert Bosch GmbH hereby grants to licensee the non-exclusive and 
 * non-transferable right to use the software and accompanying 
 * documentation for the sole purpose of testing designs delivered by 
 * Robert Bosch GmbH to licensee.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND WITHOUT ANY WARRANTY,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE, TITLE AND 
 * NONINFRINGEMENT. IN NO EVENT SHALL ROBERT BOSCH GMBH BE LIABLE FOR
 * ANY DIRECT OR INDIRECT DAMAGES, INJURY OR OTHER LIABILITY, ARISING
 * OUT OR IN CONNECTION WITH THE SOFTWARE. ROBERT BOSCH GMBH SHALL
 * NOT BE LIABLE FOR THE ACCURACY OF THE DESIGNS TESTED WITH THE AID
 * OF THE SOFTWARE.
*/

#ifndef IFS_CONSTRAINTS_H
#define IFS_CONSTRAINTS_H

#include <scv.h>
#include <stdarg.h>
#include <map>
#include <string>
#include <sstream>
#include "IFS_Message.h"
#include "IFS_Utils.h"
#include "IFS_OutputHandler.h"
#include "IFS_SmartPtr.h"

// forward declaration due to problems with some scv header files
void _scv_pop_constraint();

namespace IFS {

//--------------------------------------------------------------------------------------
//IFS_ConstraintBase
//--------------------------------------------------------------------------------------
  class IFS_ConstraintBase : public scv_constraint_base
  {
    public:
      SCV_CONSTRAINT_CTOR(IFS_ConstraintBase)
      {
        end_ptr.disable_randomization();
        end_ptr.set_name("end");
      }

      /// \brief
      /// store names of ifs_smart_ptr objects
      /// \param ifs_ptr = pointer to ifs_smart_ptr object
      /// \param ifs_ptr_name = name of the smart pointer
      /// \param default_value = a default value
      template <typename T>
      void registerConstraint(ifs_smart_ptr<T> *const ifs_ptr, std::string ifs_ptr_name,T default_value) throw (IFS_ErrorMessage);

      /// \brief
      /// used to change a constraint value from cmd-file
      /// \param ifs_ptr = a smart pointer
      /// \param parameters = list of parameters to identify ifs_smart_ptr member function
      template <typename T>
      void adapt_constraint(T ifs_ptr, std::list<std::string> parameters) throw (IFS_ErrorMessage);

    private:
      /// \brief
      /// used to check, if parameter contains a range definition
      /// \param exprList
      /// \return Returns true, if the expression contains a range
      bool isRange(const std::list<std::string>& exprList) const;

      /// \brief
      /// used to check, if parameter contains a probability
      /// \param exprList
      /// \return Returns true, if the expression contains a probability
      bool containsProbability(const std::list<std::string>& exprList) const;

      /// \brief
      /// used to parse a KEEP_LIST-command from cmd-file
      /// \param ifs_ptr = the ifs_smart_ptr
      /// \param function_name = name of requested function
      /// \param keepExpr = contains new constraint values
      /// \return Returns true, if the statement is well-formed
      template <typename T>
      bool parseKeepListStatement(T ifs_ptr, std::string function_name, std::list<std::string> keepExpr) throw (IFS_ErrorMessage);

      /// \brief
      /// used to parse a KEEP_ONLY or KEEP_OUT command from cmd-file
      /// \param ifs_ptr = the ifs_smart_ptr
      /// \param function_name = name of requested function
      /// \param keepExpr = contains new constraint values
      /// \return Returns true, if the statement is well-formed
      template <typename T>
      bool parseKeepStatement(T ifs_ptr, std::string function_name, std::list<std::string> keepExpr) throw (IFS_ErrorMessage);

      /// \brief
      /// used to parse a DISTRIBUTION command from cmd-file
      /// \param ifs_ptr = the ifs_smart_ptr
      /// \param distExpr = contains new constraint values
      /// \return Returns true, if the statement is well-formed
      template <typename T>
      bool parseDistributionStatement(T ifs_ptr, std::list<std::string> distExpr) throw (IFS_ErrorMessage);

    private:
      /// \brief
      /// used to indicate the end of a variable list of smart pointers
      /// \return smart pointer that indicates the end of a variable list of smart pointers
      scv_smart_ptr<bool>* end();

      /// indicates the end of a variable list
      ifs_smart_ptr<bool> end_ptr;

      /// store registered constraints (smart pointers)
      map<std::string,std::string> m_ConstraintMapping;
  };
}//IFS

//------------------------------------------------------------------------------
//------------------------------------------------------------------------------
// Templates
//------------------------------------------------------------------------------
//------------------------------------------------------------------------------

//------------------------------------------------------------------------------
//IFS_ConstraintBase
//------------------------------------------------------------------------------
template <typename T>
void
IFS::IFS_ConstraintBase::adapt_constraint(T ifs_ptr, list<std::string> parameters) throw (IFS_ErrorMessage)
{
  if(parameters.empty())
  {
    throw IFS_ErrorMessage("Too few tokens!",
          IFS_Message::Module,
          "ADAPT command illegal!");
  }
  std::string function_name = parameters.front();
  parameters.pop_front();

  unsigned int temp_value_1;

  if(function_name == "DISABLE")
  {
    switch(parameters.size())
    {
      case 0:
        ifs_ptr.disable_randomization();
        ifs_ptr.write(ifs_ptr.default_value);
      break;
      case 1:
        temp_value_1 = IFS::StringToUInt(parameters.front());
        parameters.pop_front();
        ifs_ptr.disable_randomization();
        ifs_ptr.write((typeof(ifs_ptr.read()))temp_value_1);
      break;
      default:
        throw IFS_ErrorMessage("DISABLE: Parameter mismatch!",
              IFS_Message::Module,
              "");
    } //switch(parameters.size())
    std::stringstream s1;
    s1 << "Disabled Randomization! Set " << ifs_ptr.get_name() << " to: " << ifs_ptr.read();
    g_IfsOut << IFS_Message(s1.str(),
          IFS_Message::Module,
          "" ,
          IFS_Message::Info);
    s1.str(""); s1.clear();
  }
  else if (function_name == "KEEP_ONLY")
  {
    parseKeepStatement(ifs_ptr,function_name,parameters);
    parameters.clear();
  }
  else if (function_name == "KEEP_OUT")
  {
    parseKeepStatement(ifs_ptr,function_name,parameters);
    parameters.clear();
  }
  else if (function_name == "ENABLE")
  {
    ifs_ptr.enable_randomization();
    g_IfsOut << IFS_Message("Enabled Randomization:" + ifs_ptr.get_name(),
          IFS_Message::Module,
          "" ,
          IFS_Message::Info);
  }
  else if (function_name == "CLEAR")
  {
    ifs_ptr.clear();
    g_IfsOut << IFS_Message("Cleared: " + ifs_ptr.get_name(),
          IFS_Message::Module,
          "" ,
          IFS_Message::Info);
  }
  else if (function_name == "DISTRIBUTION")
  {
    parseDistributionStatement(ifs_ptr,parameters);
    parameters.clear();
  }
  else if (function_name == "KEEP_ONLY_LIST")
  {
    parseKeepListStatement(ifs_ptr,function_name,parameters);
    parameters.clear();
  }
  else if (function_name == "KEEP_OUT_LIST")
  {
    parseKeepListStatement(ifs_ptr,function_name,parameters);
    parameters.clear();
  }
  else
  {
    throw IFS_ErrorMessage("Invalid Token!",
          IFS::IFS_Message::Module,
          "ADAPT command illegal!");
  }
}

//------------------------------------------------------------------------------

template <typename T>
bool IFS::IFS_ConstraintBase::parseKeepListStatement(T ifs_ptr, std::string function_name, list<std::string> keepExpr) throw (IFS_ErrorMessage)
{
  std::list<std::string>::iterator iter         = keepExpr.begin();
  std::list<std::string>::iterator tempIterator = keepExpr.begin();
  std::list<std::string>::iterator endIter      = keepExpr.end();

  const std::string requestStart  = "{";
  const std::string requestEnd    = "}";

  std::string temp_str_1;
  std::string temp_str_2;

  unsigned int temp_value_1;
  unsigned int temp_value_2;

  if((*iter) == requestStart)
  {
    if(keepExpr.back() != requestEnd)
    {
      throw IFS_ErrorMessage("Malformed KEEP_LIST!",
            IFS_Message::Module,
            "");
    }
    while (((*iter) != requestEnd) && (iter != endIter))
    {
      iter++;
      if(iter == endIter)
      {
        IFS_ErrorMessage("Malformed KEEP_LIST!",
            IFS_Message::Module,
            "");
      }
      temp_str_1 = *iter;
      //set iter to "," or "}"
      iter++;
      if((*iter != "}") && (*iter != ","))
      {
        throw IFS_ErrorMessage("Malformed KEEP_LIST! Token: " + *iter,
              IFS_Message::Module,
              "");
      }
      //conversion & update distribution
      temp_value_1 = IFS::StringToUInt(temp_str_1);
      if (function_name == "KEEP_ONLY_LIST")
      {
        ifs_ptr.update_keep_only_list((typeof(ifs_ptr.read()))temp_value_1);
        g_IfsOut << IFS_Message( "KEEP_ONLY_LIST : Updated " + ifs_ptr.get_name() + " with " + temp_value_1,
            IFS_Message::Module,
            "" ,
              IFS_Message::Info);
      }
      else if (function_name == "KEEP_OUT_LIST")
      {
        ifs_ptr.update_keep_out_list((typeof(ifs_ptr.read()))temp_value_1);
        g_IfsOut << IFS_Message( "KEEP_OUT_LIST : Updated " + ifs_ptr.get_name() + " with " + temp_value_1,
            IFS_Message::Module,
            "" ,
            IFS_Message::Info);
      }
      else
      {
        IFS_ErrorMessage("Malformed KEEP_LIST!",
            IFS_Message::Module,
            "");
      }
    }//while
  }
  else
  {
    IFS_ErrorMessage("Malformed KEEP_LIST!",
        IFS_Message::Module,
        "");
  } //if((*iter) == requestStart)
  return true;
}

//------------------------------------------------------------------------------

template <typename T>
bool IFS::IFS_ConstraintBase::parseKeepStatement(T ifs_ptr, std::string function_name, list<std::string> keepExpr) throw (IFS_ErrorMessage)
{
  list<std::string>::iterator iter         = keepExpr.begin();
  list<std::string>::iterator tempIterator = keepExpr.begin();
  list<std::string>::iterator endIter      = keepExpr.end();

  const std::string requestStart  = "{";
  const std::string requestEnd    = "}";
  const std::string range         = ":";

  std::string temp_str_1;
  std::string temp_str_2;

  unsigned int temp_value_1;
  unsigned int temp_value_2;

  if((*iter) == requestStart)
  {
    if(keepExpr.back() != requestEnd)
    {
      throw IFS_ErrorMessage("Malformed KEEP!",
            IFS_Message::Utils,
            "");
    }
    while (((*iter) != requestEnd) && (iter != endIter))
    {
      iter++;
      if(iter == endIter)
      {
        throw IFS_ErrorMessage("Malformed KEEP!",
              IFS_Message::Utils,
              "");
      }
      if(isRange(keepExpr))
      {
        tempIterator = iter;
        tempIterator++;
        if(tempIterator == endIter)
        {
          throw IFS_ErrorMessage("Malformed KEEP!",
                IFS_Message::Utils,
                "");
        }
        if((*tempIterator) == range)
        {
          temp_str_1 = *iter;
          iter++;
          iter++;
          if(iter == endIter)
          {
            throw IFS_ErrorMessage("Malformed KEEP!",
                  IFS_Message::Utils,
                  "");
          }
          temp_str_2 = *iter;
          //set iter to "," "}"
          iter++;
          //conversion & update keep
          temp_value_1 = IFS::StringToUInt(temp_str_1);
          temp_value_2 = IFS::StringToUInt(temp_str_2);
          if (function_name == "KEEP_ONLY")
          {
            ifs_ptr.update_keep_only((typeof(ifs_ptr.read()))temp_value_1,(typeof(ifs_ptr.read()))temp_value_2);
            g_IfsOut << IFS_Message( "KEEP_ONLY : Updated " + ifs_ptr.get_name() + " with " + temp_value_1 + " : " + temp_value_2,
              IFS_Message::Utils,
              "" ,
              IFS_Message::Info);
          }
          else if (function_name == "KEEP_OUT")
          {
            ifs_ptr.update_keep_out((typeof(ifs_ptr.read()))temp_value_1,(typeof(ifs_ptr.read()))temp_value_2);
            g_IfsOut << IFS_Message( "KEEP_OUT : Updated " + ifs_ptr.get_name() + " with " + temp_value_1 + " : "  + temp_value_2,
              IFS_Message::Utils,
              "" ,
              IFS_Message::Info);
          }
          else
          {
            throw IFS_ErrorMessage("Malformed KEEP!",
              IFS_Message::Utils,
              "");
          }
        }
        else
        {
          throw IFS_ErrorMessage("Malformed KEEP!",
            IFS_Message::Utils,
            "");

        }//if((*tempIterator) == range)
      }
      else
      {
        tempIterator = iter;
        tempIterator++;
        if(tempIterator == endIter)
        {
          throw IFS_ErrorMessage("Malformed KEEP!",
              IFS_Message::Utils,
              "");
        }
        if((*tempIterator) == requestEnd)
        {
          temp_str_1 = *iter;
          //set iter to "," "}"
          iter++;
          //conversion & update keep
          temp_value_1 = IFS::StringToUInt(temp_str_1);
          if (function_name == "KEEP_ONLY")
          {
            ifs_ptr.update_keep_only((typeof(ifs_ptr.read()))temp_value_1);
            g_IfsOut << IFS_Message( "KEEP_ONLY : Updated " + ifs_ptr.get_name() + " with " + temp_value_1,
              IFS_Message::Utils,
              "" ,
              IFS_Message::Info);
          }
          if (function_name == "KEEP_OUT")
          {
            ifs_ptr.update_keep_out((typeof(ifs_ptr.read()))temp_value_1);
            g_IfsOut << IFS_Message( "KEEP_OUT : Updated " + ifs_ptr.get_name() + " with " + temp_value_1,
                  IFS_Message::Utils,
                  "" ,
                  IFS_Message::Info);
          }
        }
        else
        {
          IFS_ErrorMessage("Malformed KEEP!",
              IFS_Message::Utils,
              "");
        } //if((*tempIterator) == requestEnd)
      } //if(isRange(keepExpr))
    } //while
  }
  else
  {
    throw IFS_ErrorMessage("Malformed KEEP!",
          IFS_Message::Utils,
          "");
  } //if((*iter) == requestStart)
  return true;
}

//------------------------------------------------------------------------------

template <typename T>
bool IFS::IFS_ConstraintBase::parseDistributionStatement(T ifs_ptr, list<std::string> distExpr) throw (IFS_ErrorMessage)
{
  list<std::string>::iterator iter         = distExpr.begin();
  list<std::string>::iterator tempIterator = distExpr.begin();
  list<std::string>::iterator endIter      = distExpr.end();

  const std::string requestStart  = "{";
  const std::string requestEnd    = "}";
  const std::string range         = ":";
  const std::string probability   = "%";

  std::string temp_str_1 = "";
  std::string temp_str_2 = "";
  std::string temp_str_3 = "";

  unsigned int temp_value_1 = 0;
  unsigned int temp_value_2 = 0;
  unsigned int temp_value_3 = 0;

  if((*iter) == requestStart)
  {
    if(distExpr.back() != requestEnd)
    {
      throw IFS_ErrorMessage("Malformed Distribution!",
            IFS_Message::Utils,
            "");
    }

    while (((*iter) != requestEnd) && (iter != endIter))
    {
      iter++;
      if(iter == endIter)
      {
        throw IFS_ErrorMessage("Malformed Distribution!",
              IFS_Message::Utils,
              "");
      }
      if(isRange(distExpr))
      {
        tempIterator = iter;
        tempIterator++;
        if(tempIterator == endIter)
        {
          throw IFS_ErrorMessage("Malformed Distribution!",
              IFS_Message::Utils,
              "");
        }
        if((*tempIterator) == range)
        {
          tempIterator = iter;
          tempIterator++;
          if(tempIterator == endIter)
          {
            throw IFS_ErrorMessage("Malformed Distribution!",
                IFS_Message::Utils,
                "");
          }
          tempIterator++;
          if(tempIterator == endIter)
          {
            throw IFS_ErrorMessage("Malformed Distribution!",
              IFS_Message::Utils,
              "");
          }
          tempIterator++;
          if(tempIterator == endIter)
          {
            throw IFS_ErrorMessage("Malformed Distribution!",
              IFS_Message::Utils,
              "");
          }
          if((*tempIterator).substr(((*tempIterator).length()-1),1)  == probability)
          {
            temp_str_1 = *iter;
            iter++;
            if(iter == endIter)
            {
              throw IFS_ErrorMessage("Malformed Distribution!",
                  IFS_Message::Utils,
                  "");
            }
            iter++;
            if(iter == endIter)
            {
              throw IFS_ErrorMessage("Malformed Distribution!",
                  IFS_Message::Utils,
                  "");
            }
            temp_str_2 = *iter;
            iter++;
            if(iter == endIter)
            {
              throw IFS_ErrorMessage("Malformed Distribution!",
                  IFS_Message::Utils,
                  "");
            }
            temp_str_3 = (*iter).substr(0,((*iter).length()-1));
            //set iter to "," or "}"
            iter = tempIterator;
            iter++;
            if(iter == endIter)
            {
              throw IFS_ErrorMessage("Malformed Distribution!",
                  IFS_Message::Utils,
                  "");
            }
            //conversion & update distribution
            temp_value_1 = IFS::StringToUInt(temp_str_1);
            temp_value_2 = IFS::StringToUInt(temp_str_2);
            temp_value_3 = IFS::StringToUInt(temp_str_3);
            ifs_ptr.update_distribution((typeof(ifs_ptr.read()))temp_value_1,(typeof(ifs_ptr.read()))temp_value_2,(typeof(ifs_ptr.read()))temp_value_3);
            g_IfsOut << IFS_Message( "DISTRIBUTION : Updated " + ifs_ptr.get_name() + " with " + temp_value_1 + " : " + temp_value_2 + " " + temp_value_3 + "%",
                IFS_Message::Utils,
                "" ,
                IFS_Message::Info);
          }
          else
          {
            throw IFS_ErrorMessage("Malformed Distribution!",
                IFS_Message::Utils,
                "");
          }//if((*tempIterator) == probability)
        }
        else
        {
          throw IFS_ErrorMessage("Malformed Distribution!",
              IFS_Message::Utils,
              "");
        }//if((*tempIterator) == range)
      }
      else
      {
        if (containsProbability(distExpr))
        {
          tempIterator = iter;
          tempIterator++;
          if(tempIterator == endIter)
          {
            throw IFS_ErrorMessage("Malformed Distribution!",
              IFS_Message::Utils,
              "");
          }
          if((*tempIterator).substr(((*tempIterator).length()-1),1)  == probability)
          {
            temp_str_1 = *iter;
            iter++;
            if(iter == endIter)
            {
              throw IFS_ErrorMessage("Malformed Distribution!",
                IFS_Message::Utils,
                "");
            }
            temp_str_2 = (*iter).substr(0,((*iter).length()-1));
            //set iter to "," or "}"
            iter++;
            if(iter == endIter)
            {
              throw IFS_ErrorMessage("Malformed Distribution!",
                  IFS_Message::Utils,
                  "");
            }
            //conversion & update distribution
            temp_value_1 = IFS::StringToUInt(temp_str_1);
            temp_value_2 = IFS::StringToUInt(temp_str_2);
            ifs_ptr.update_distribution((typeof(ifs_ptr.read()))temp_value_1, (typeof(ifs_ptr.read()))temp_value_2);
            g_IfsOut << IFS_Message( "DISTRIBUTION : Updated " + ifs_ptr.get_name() + " with " + temp_value_1 + " " + temp_value_2 + "%",
                IFS_Message::Utils,
                "" ,
                IFS_Message::Info);
          }
          else
          {
            throw IFS_ErrorMessage("Malformed Distribution!",
                IFS_Message::Utils,
                "");
          }//if tempIterator == probability
        }
        else
        {
          temp_str_1 = *iter;
          iter++;
          if(iter == endIter)
          {
            throw IFS_ErrorMessage("Malformed Distribution!",
                IFS_Message::Utils,
                "");
          }
          //conversion & update distribution
          temp_value_1 = IFS::StringToUInt(temp_str_1);
          ifs_ptr.update_distribution((typeof(ifs_ptr.read()))temp_value_1);
          g_IfsOut << IFS_Message( "DISTRIBUTION : Updated " + ifs_ptr.get_name() + " with " + temp_value_1,
              IFS_Message::Utils,
              "" ,
              IFS_Message::Info);
        }//if (containsProbability(distExpr))
      }//if(isRange(distExpr))
    }//while (((*iter) != requestEnd) && (iter != endIter))
  }
  else
  {
    throw IFS_ErrorMessage("Malformed Distribution!",
        IFS_Message::Utils,
        "");
  }//if((*iter) == requestStart)
  return true;
}

//------------------------------------------------------------------------------

template <typename T>
void IFS::IFS_ConstraintBase::registerConstraint(ifs_smart_ptr<T> *const ifs_ptr, std::string ifs_ptr_name,T default_value) throw (IFS_ErrorMessage)
{
  if( m_ConstraintMapping.find(ifs_ptr_name) != m_ConstraintMapping.end() )
  {
    throw IFS_ErrorMessage("Constraint already registered: " +
        ifs_ptr_name + " within: " + this->get_name(),
        IFS_Message::Utils,
        "");
  }
  else
  {
    m_ConstraintMapping[ifs_ptr_name] = ifs_ptr_name;
    ifs_ptr->set_name(ifs_ptr_name);
    ifs_ptr->default_value = default_value;
    g_IfsOut << IFS_Message( "DEFAULT VALUE for: " + ifs_ptr->get_name() + " is: " + ifs_ptr->default_value,
          IFS_Message::Utils,
          "" ,
          IFS_Message::Info);
  }
}

#endif

