/*
 *  (C) Robert Bosch GmbH 2008. All rights reserved, also regarding any
 *  disposal, exploitation, reproduction, editing, distribution, as
 *  well as in the event of application for industrial property rights.
 *
 *  Contact: AE/NE1-IFS <vAENE1.IFS@de.bosch.com>
 *
 * Disclaimer:
 * This Software is licensed under the following conditions. You may 
 * not use this file except in compliance with the following conditions:
 *
 * Robert Bosch GmbH hereby grants to licensee the non-exclusive and 
 * non-transferable right to use the software and accompanying 
 * documentation for the sole purpose of testing designs delivered by 
 * Robert Bosch GmbH to licensee.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND WITHOUT ANY WARRANTY,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE, TITLE AND 
 * NONINFRINGEMENT. IN NO EVENT SHALL ROBERT BOSCH GMBH BE LIABLE FOR
 * ANY DIRECT OR INDIRECT DAMAGES, INJURY OR OTHER LIABILITY, ARISING
 * OUT OR IN CONNECTION WITH THE SOFTWARE. ROBERT BOSCH GMBH SHALL
 * NOT BE LIABLE FOR THE ACCURACY OF THE DESIGNS TESTED WITH THE AID
 * OF THE SOFTWARE.
*/

#ifndef IFS_CONTROL_H
#define IFS_CONTROL_H

// STL-components
#include <list>
#include <string>

// IFS components
#include "IFS_OutputHandler.h"

namespace IFS
{
  /// \note This is defined in IFS_ModuleBase.h, too. However, spending
  ///       an extra header file for this seems to be overkill.
  typedef unsigned int InterruptID;

  class IFS_ModuleBase;

  /// Contains the implementation of the controller.
  /** The implementation itself does not need to be visible for the user.
    */
  namespace Impl 
  {
    class IFS_Control;
  }

  /// Controls the execution of the IFS script.
  /** This class is the access-point for the IFS-modules and also
    * controls the modules' accesses to their commands.
    * \note Actually this is a wrapper for IFS::Impl::IFS_Control in order
    *       to keep the \#include's seen by the user minimal -- and especially
    *       avoid conflicting <systemc.h> includes with different SC_INCLUDE_FX
    *       settings. (The fixed point types are needed for the expression 
    *       evaluation, but this functionality is not directly accessible to the
    *       user).
    *       An inheritance-based approach would be nicer (as opposed to manually
    *       forward the method calls), but would break the existing IFS usage 
    *       mechanism, which allows to instantiate an IFS_Control object via its
    *       default constructor.
    */        
  class IFS_Control
  {
    public:

      /// Constructor
      IFS_Control();

      /// Destructor
      ~IFS_Control();

      // Disables the static script checking 
      void disableStaticScriptChecking();

      /// \brief
      /// Pass coverage information to a specific command file variable
      /// 
      /// \param var_name = name of command file variable and coverage signal
      /// \param var_value = assigned value
      void passCoverageInfoToCmdFileVariable(const std::string& var_name, const std::string& var_value);

      /// Registers the specified module for usage with the controller.
      void registerModule( IFS_ModuleBase &newModule );

      /// \brief
      ///   Load the IFS script to be executed
      /// \note The execution of the script starts with the simulation.
      void loadScript( const std::string &scriptFileName );

      /// \brief
      /// Looks in the loaded script for the next command for the module
      /// specified in the parameter and executes the appropriate method of
      /// said module.
      ///
      /// This method is called from within the command loop of IFS_ModuleBase
      /// and performs the execution of the next command for the requesting
      /// module.
      /// \return Returns true, if a command had been found, else false.
      bool executeNextCommand( IFS_ModuleBase *pRequestingModule );

      /// Displays data/statistics about the simulation-run so far
      /** Prints number of warnings and errors
        */      
      void report() const;
      
      /// Notify the controller about interrupts that occured.
      /** \note This method has to be called immediately when the interrupt 
        *       occurs. It will store the interrupt request(s) in the 
        *       executionState of each module.
        * \param occurredInterrupts list of simultaneously occurred interrupts.
        */
      void interruptsOccurred(const std::list<InterruptID>& occurredInterrupts);

      /// The minimum severity for messages to be displayed
      void setReportLevel( const IFS_Message::Severity &minSeverity );
      
      /// A file (named Trace_sc.log) is written, which contains
      /// all outputs of the g_IfsOut object.
      void writeTraceLog(const std::string& fileName = "Trace_sc.log");
      
      /// Enables the debug mode for all registered IFS-Modules
      /// This method must be called after all modules have been registered 
      void enableDebugMode();
      
      /// Enables the debug mode for the given IFS-Modules in the list.
      /// The list (value) includes the names of the IFS-Modules.
      /// This method must be called after all modules have been registered.
      void enableDebugMode(const std::list<std::string>& value);

      /// The stream handle for log messages etc
      IFS_OutputHandler& getOutputHandler();

    private:
      /// The "real" controller
      ::IFS::Impl::IFS_Control    *m_Implementation;
  };
}

#endif

