/*
 *  (C) Robert Bosch GmbH 2008. All rights reserved, also regarding any
 *  disposal, exploitation, reproduction, editing, distribution, as
 *  well as in the event of application for industrial property rights.
 *
 *  Contact: AE/NE1-IFS <vAENE1.IFS@de.bosch.com>
 *
 * Disclaimer:
 * This Software is licensed under the following conditions. You may 
 * not use this file except in compliance with the following conditions:
 *
 * Robert Bosch GmbH hereby grants to licensee the non-exclusive and 
 * non-transferable right to use the software and accompanying 
 * documentation for the sole purpose of testing designs delivered by 
 * Robert Bosch GmbH to licensee.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND WITHOUT ANY WARRANTY,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE, TITLE AND 
 * NONINFRINGEMENT. IN NO EVENT SHALL ROBERT BOSCH GMBH BE LIABLE FOR
 * ANY DIRECT OR INDIRECT DAMAGES, INJURY OR OTHER LIABILITY, ARISING
 * OUT OR IN CONNECTION WITH THE SOFTWARE. ROBERT BOSCH GMBH SHALL
 * NOT BE LIABLE FOR THE ACCURACY OF THE DESIGNS TESTED WITH THE AID
 * OF THE SOFTWARE.
*/

#ifndef IFS_CONTROL_IMPL_H
#define IFS_CONTROL_IMPL_H

// STL-components
#include <list>
#include <map>
#include <string>
#include <stack>

#include "IFS_Control.h"


#define SC_INCLUDE_FX
#include <systemc.h>

#include "IFS_Script.h"
#include "IFS_Scanner.h"

using sc_core::sc_time;

namespace IFS
{

  class IFS_CommandPointerBase;

namespace Impl
{
  /// The actual implementation of IFS_Control
  class IFS_Control
  {

    public:

      /// \name User-accessible functions
      /// @{

      /// Constructor
      IFS_Control(::IFS::IFS_Control *wrapper);

      // Disables the static script checking
      void disableStaticScriptChecking();

      /// Registers the specified module for usage with the controller.
      void registerModule( IFS_ModuleBase &newModule );

      /// \brief
      /// Pass coverage information to a specific command file variable
      ///
      /// \param var_name = name of command file variable and coverage signal
      /// \param var_value = assigned value
      void passCoverageInfoToCmdFileVariable(const std::string& var_name,
                                             const std::string& var_value );

      /// \brief
      ///   Load the IFS script to be executed
      /// \note The execution of the script starts with the simulation.
      void loadScript( const std::string &scriptFileName );

      /// \brief
      /// Looks in the loaded script for the next command for the module
      /// specified in the parameter and executes the appropriate method of
      /// said module.
      ///
      /// This method is called from within the command loop of IFS_ModuleBase
      /// and performs the execution of the next command for the requesting
      /// module.
      /// \return Returns true, if a command had been found, else false.
      bool executeNextCommand( IFS_ModuleBase *pRequestingModule );

      /// Displays data/statistics about the simulation-run so far
      /** Currently the number of warnings and errors are reported
        */
      void report() const;

      /// Determines that a file (named trace.log) is written, which contains
      /// all outputs of the g_IfsOut object.
      void writeTraceLog(const std::string& fileName);

      /// Notify the controller about interrupts that occured.
      /** \note This method has to be called immediately when the interrupt
        *       occurs. It will store the interrupt request(s) in the
        *       executionState of each module.
        * \param occurredInterrupts list of simultaneously occurred interrupts.
        */
      void interruptsOccurred(const std::list<InterruptID>& occurredInterrupts);

      /// Enables the debug mode for all registered IFS-Modules
      /// This method must be called after all modules have been registered
      void enableDebugMode();

      /// Enables the debug mode for the given IFS-Modules in the list.
      /// The list (value) includes the names of the IFS-Modules.
      /// This method must be called after all modules have been registered.
      void enableDebugMode(const std::list<std::string>& value);

      /// The minimum severity for messages to be displayed
      void setReportLevel( const IFS_Message::Severity &minSeverity );

      /// @}

    private:

      /// Pointer to an instruction (i.e. line) in the script
      typedef IFS_Script::const_iterator InstructionPointer;


      /// \defgroup ModuleStateTypes Controller-internal types related to an IFS module's state
      /// @{

      /// Information needed for executing a loop
      /** \note It would be sufficient to store the value of the loop variable,
        *       only. The others could be obtained by rescanning the script.
        *       However, it is more sensible to evaluate the \#loop statement
        *       once (for each module), so executing it can be seen as creating
        *       one instance of the loop.
        */
      struct LoopContext
      {
        /// current value of the loop variable
        int                    m_CurrentValue;
        /// first line after the \#loop
        InstructionPointer     m_LoopBodyBegin;
        /// name of the iteration variable (without '\#')
        std::string            m_IterationVariable;
        /// name of the enumeration variable (without '\#')
        std::string            m_EnumerationVariable;
        /// value of the loop variable during first iteration
        int                    m_IntervalStart;
        /// value of the loop variable during last iteration
        int                    m_IntervalEnd;
        /// sequence of enumeration values;
        std::list<std::string> m_EnumerationValues;
      };

      /// Can store a module's local execution state
      /** A module can have more than one ExecutionContext (when executing
        * an interrupt handler), but only one active ExecutionContext.
        * The active ExecutionContext is given by the top element of
        * IFS::Impl::IFS_Control::ModuleExecutionState::m_ExecutionContextStack.
        */
      struct ExecutionContext
      {
        /// Constructor
        ExecutionContext() : m_ExecutingWaitSimtime(false),
                             m_ExecutingSync(false) {}

        /// Points to the next instruction to be executed
        /** Points to \c m_Script.end() when there is no next instruction
          */
        InstructionPointer  m_InstructionPointer;

        /// values of (nested) loop variables
        /** .back() is innermost loop
          * A list is chosen instead of a stack due to efficency reasons
          * (you cannot iterate over the elements of a stack without
          *  changing it)
          */
        std::list<LoopContext>  m_LoopContextStack;

        /// true => we are currently executing a WAIT simtime command
        /** \note WAIT deltime is handled by the same methods as WAIT
          *       simtime, hence this flag also indicates a WAIT deltime
          */
        bool                m_ExecutingWaitSimtime;
        /// \brief
        /// Ending time of a WAIT simtime/deltime command
        /// Only meaningful if m_ExecutingWaitSimtime == true
        sc_time             m_WaitSimtimeEnd;
        /// \brief
        /// The position of the currently executed wait statement
        /// Only meaningful if m_ExecutingWaitSimtime == true
        InstructionPointer  m_WaitPosition;
        /// true => we are currently executing a SYNC command
        bool                m_ExecutingSync;
        /// \brief
        /// the SYNC where we are waiting (in case of multiple SYNCs)
        /// only meaningful if m_ExecutingSync == true
        InstructionPointer  m_SyncPosition;
      };

      /// Can store a module's global execution state
      /** This structure holds all information which reflect the current
        * state of execution.
        */
      struct ModuleExecutionState
      {
        /// default ctor for proper initialisation
        ModuleExecutionState() : m_ExecutingInterruptHandler(false)
                                 {}

        /// true if we are currently executing an interrupt handler
        bool                     m_ExecutingInterruptHandler;

        /// If an interrupt is currently handled, this variable holds its ID
        InterruptID              m_ExecutedInterruptHandler;

        /// Interrupt requests that have to be handled (.front() is oldest)
        std::list<InterruptID>   m_InterruptRequestList;

        /// Holds the execution-contexts for suspended interrupts (\#break)
        std::map< InterruptID, ExecutionContext > m_SuspendedInterrupts;

        /// \brief Store old execution state during interrupt
        /** \note stacksize is 1 or 2 (no nested interrupts)
          */
        std::stack<ExecutionContext>  m_ExecutionContextStack;

      };

      /// Contains (per module) information as needed by the controller
      /** We could also put this in the ModuleBase instead of having a map here.
        * However, currently it seems more sensible not to put it in the
        * ModuleBase, because the ModuleBase itself does not need these
        * attributes and they should not be visible to the user either.
        */
      struct ModuleInfo
      {
        ModuleInfo(std::map< std::string,
                   IFS::IFS_CommandPointerBase*> const * commandMapping=NULL)
                     : m_CommandMapping(commandMapping),
                       m_pResumeCommandLoopEvent(NULL),
                       m_CommandLoopSuspended(false),
                       m_DebugState(0)
                       {}

        /// The module's global execution state.
        ModuleExecutionState m_ExecutionState;

        /// command mapping ( command name -> method pointer)
        std::map< std::string, IFS::IFS_CommandPointerBase*> const * m_CommandMapping;

        /// \brief
        /// Needed to resume the (waiting) command loop.
        /// \note sc_event has its copy constructor disabled, hence the pointer
        /// \note One might expect this as part of the ExecutionContext.
        ///       However, since there is only one command loop thread per
        ///       module, it is sufficient to store a single event per module.
        sc_event *m_pResumeCommandLoopEvent;

        /// \brief
        /// Indicates that the command loop is waiting for the
        /// m_pResumeCommandLoopEvent
        bool m_CommandLoopSuspended;

        // stores the debug state
        int m_DebugState;

        /// stores the breakpoints for the debug mode
        std::map<std::string, std::list<unsigned int> > breakPoints;

        /// \brief
        /// Waits until m_WaitSimtimeEnd (of the current executionContext)
        //  or until m_InterruptEvent is notified.
        /// \note Control should be passed back to the mainLoop of the
        ///       requesting module immediately after this method returns.
        void WaitSimtime();
      };

      /// Information associated to an interrupt
      struct InterruptInfo
      {
        /// default ctor
        InterruptInfo() : m_IsEnabled(false), m_Priority(0) {}

        /// false => interrupt will completely be ignored
        bool                      m_IsEnabled;
        /// current priority (higher value means higher priority)
        int                       m_Priority;
        /// points to the first instruction of the handler
        InstructionPointer        m_Handler;
      };

      /// @}

      struct HasName;


      // Attributes:
      // -----------

      /// The script that will be executed
      IFS_Script m_Script;

      typedef std::map< const IFS_ModuleBase *, ModuleInfo > ModuleRegister ;
      /// Maps each registered module to its corresponding ModuleInfo
      ModuleRegister m_RegisteredModules;

      /// Get InterruptInfo for an interrupt via its ID
      std::map< InterruptID, InterruptInfo > m_InterruptInfo;

      /// The "environment" containing the variables.
      /** These variables are set by a predefined module command and may
        * be used in expressions.
        */
      std::map< std::string, std::string>  m_GlobalVariables;

      /// The "environment" containing the variable-arrays.
      /** These variables are set by a predefined module command and may
        * be used in expressions.
        */
      std::map< std::string, std::list<std::string> >  m_GlobalArrayVariables;

      /// Stores the information whether the simulation is to be quitted or not
      bool m_QuitSimulation;

      /// Stores the information whether the script checking is to be done or not
      bool m_ExecuteScriptChecking;

      /// \brief
      /// The default accuracy for the evaluation of expressions
      static const sc_fxtype_params s_DefaultExpressionAccuracy;

      /// \brief
      /// The accuracy for the evaluation of expressions (which are based on
      /// sc_ufix, see IFS::Expression::ArithmeticExpressionType)
      sc_fxtype_params m_ExpressionAccuracy;


      /// This controller's wrapper
      ::IFS::IFS_Control    *m_pWrapper;

      // stores which values have already been used during the executions
      static std::map<std::string, std::list<unsigned int > > m_RandomValuesWhichHaveBeenAlreadyUsed;
      /// \name Script Processing
      /// @{


      /// Calls the debug prompt
      IFS::Scanner::ScannedLine debugging(const IFS_ModuleBase *pRequestingModule, const InstructionPointer& ip, const IFS::Scanner::ScannedLine& scannedLine);

      /// \brief
      /// Perform static check of the internal script for correctness and
      /// throw errors, if necessary.
      void checkScript() const;

      /// \brief
      /// Find \#isrs and fill in the internal m_InterruptInfo structure
      /// \note Locating interrupt-handlers is done without substitutions.
      void locateInterruptHandlers();


      /// \brief Locates the initial command (in the internal script) where all
      /// command-requests of modules start from. If no such line exists
      /// an iterator to \c m_Script.end() is returned
      /// \return A constant iterator pointing to the line in the local script
      /// with the initial command.
      IFS_Script::const_iterator getInitialCommand();


      /// The next instruction to be executed.
      /** The next instruction may be the next line, an ISR or a re-executed
        * SYNC.
        *
        * \param executionState The executionState of the requesting module
        * \param moduleName     The name of the module (only for messages)
        *
        */
      InstructionPointer
      findNextInstruction( ModuleExecutionState &executionState,
                           const std::string &moduleName );


      /// Find the \c searchedStatement within a hierarchy of nested blocks
      /** For example, find a matching \#else statement in a hierarchy of
        * \#if and \#endif.
        * \param nextLine must point to the first instruction after the block
        *        begin
        * \param blockBegin is the token for a block begin, e.g. \#if
        * \param blockEnd is the token for a block end, e.g. \#endif
        * \param searchedStatement is the token we are looking for, e.g. \#else
        * \return An instruction pointer to the line with the \c searchedStatement
        *         or \c m_Script.end() if the hierarchy does not contain the
        *         \c searchedStatement
        * \note No substitutions are made during the search
        */
      InstructionPointer
      findMatchingStatement(InstructionPointer nextLine,
                            const std::string& blockBegin,
                            const std::string& blockEnd,
                            const std::string& searchedStatement) const;


      /// \brief Replace all occurrences of ${ ... } in the list
      /// \param inputList sequence of tokens
      /// \param inputLine the line where the tokens are coming from
      /// \param repetitionOnOff specifies whether the repetition is on(value = true) or not (value=false)
      /// \param l_moduleName defines the name of current module
      /// \param recursionDegree is the number of the recursion
      /// \param iterationDegree is the number of the iteration
      /// \param numberOfRandomExpressionInThisLine is the number of the random expression in the line
      /// \return \c inputList with each occurence
      ///   of ${ ... } replaced by an appropriate random value.
      std::list<IFS::Scanner::Token>
      randomise( const std::list<IFS::Scanner::Token>& inputList,
                 const IFS::Scanner::ScannedLine& inputLine,
                 bool repetitionOnOff = false,
                 const std::string& l_moduleName = "",
                 unsigned int recursionDegree = 0,
                 unsigned int iterationDegree = 0,
                 unsigned int numberOfRandomExpressionInThisLine = 0) const;

      ///\brief speciefies whether a parameter for the randomisation has to be omitted or not
      // \param key the key for the map  m_RandomValuesWhichHaveBeenAlreadyUsed
      // \param value the value which could be omitted
      // \return true if the value has to be omitted, else false
      bool exceptValueForRandomization(const std::string& key,
                                  const unsigned int value) const;


      ///\brief stores the information which values have been used during a randomisation
      /// in the member m_RandomValuesWhichHaveBeenAlreadyUsed
      /// \param key the key for the map  m_RandomValuesWhichHaveBeenAlreadyUsed
      /// \param value the value which ought to be stored
      /// \param numberOfRegardedValues the number of regarded parameters
      void storeValueForRandomization(const std::string& key,
                                    const unsigned int value,
                                    const unsigned int numberOfRegardedValues) const;

      /// \brief Check if token is of the form #Varname
      /// \param token Input token to test
      bool checkForVariable(const std::string& token) const;

      /// \brief Replace all occurences of \#\<Identifier\> by their values
      /// \param scannedLine sequence of tokens possibly containing variables
      /// \param loopContextStack Context for the substitution.
      /// \param globalVariables The map of variables
      /// \param globalArrayVariables The map of variables with more than one value (the arrays)
      /// \return true if a variable was found, else false
      bool
      substituteVariables( IFS::Scanner::ScannedLine &scannedLine,
                           const std::list<LoopContext> &loopContextStack,
                           const std::map<std::string, std::string> &globalVariables,
                           const std::map<std::string, std::list<std::string> > &globalArrayVariables) const;

      /// @}


      /// \defgroup ParseAndExecuteMethods Parse and Execute Methods of the Controller
      /// @{


      /// \brief
      /// Representation of an assignment statement needed to pass from the
      /// parse method to the execute method
      struct AssignmentStatement
      {
        /// Left hand side, i.e. the variable name
        std::string m_lhs;
        /// Right hand side, i.e. the value
        std::list<std::string> m_rhs;
      };

       /// Parse the ASSIGN command
      /** \pre scannedLine.m_LineContent.front() == \c ASSIGN
       */
      AssignmentStatement
      parseAssignmentStatement ( Scanner::ScannedLine scannedLine,const ModuleInfo & moduleInfo,
      const std::map< std::string, std::string>& l_GlobalVariables,
      const std::map< std::string, std::list<std::string> >& l_GlobalArrayVariables) const;

       /// Parse the GETVALUE command
      /** \pre scannedLine.m_LineContent.front() == \c ASSIGN
       */
      AssignmentStatement
      parseGetValueStatement( Scanner::ScannedLine scannedLine,const IFS::IFS_ModuleBase* pRequestingModule ) const;


      /// Execute the given assignment statement
      void
      executeAssignmentStatement( const AssignmentStatement &,const ModuleInfo & moduleInfo);


      /// \brief
      /// The different types of module commands (plus \c NoCommand in order
      /// to indicate that the current line did not contain a ModuleCommand).
      /** The distinction is necessary during the execution of the command loop,
        * because control is passed back to the requesting module only when
        * a real module command was executed. Hence controller commands like
        * loop handling is fully transparent for the requesting module.
        */
      enum ModuleCommandKind
      {
        NoCommand,
        UserCommand,
        PredefinedCommand,
        LegacyCommand
      };

      /// Execute a ControlCommand (i.e. a command starting with \c IFS or \c ALL)
      ModuleCommandKind
      executeControlCommand(
        Scanner::ScannedLine scannedLine,
        IFS_ModuleBase *pRequestingModule,
        InstructionPointer instructionPointer,
        const ModuleInfo & moduleInfo );


      /// The parsed version of an \#if statement
      struct IfStatement
      {
        /// The condition following the \#if
        std::string         m_Condition;
        /// The first line of the then-body
        InstructionPointer  m_ThenBegin;

        /// The parsed version of a \#elseif statement
        struct ElseIfLabel
        {

          /// The label
          std::string         m_Value;
          /// The line following the \#elseif statement
          InstructionPointer  m_BlockBegin;
        };

        /// All \#elseif lables belonging to this \#if
        std::list<ElseIfLabel>  m_ElseIfLabels;

        /// The first line of the else-body (or \c m_Script.end() if there is no \#else)
        InstructionPointer  m_ElseBegin;
        /// The line following the \#endif
        InstructionPointer  m_EndIf;
      };

      /// Parse the \#if statement
      /** \pre scannedLine.front() == \#if
        */
      IfStatement
      parseIfStatement( Scanner::ScannedLine const& scannedLine,ModuleInfo & moduleInfo,
                        InstructionPointer nextLine ) const;

      /// Execute \c \#if statement
      void
      executeIfStatement(ModuleInfo & moduleInfo,
                         IfStatement const& ifStatement);

      /// Execute \#else statement, i.e., jump to the corresponding \#endif
      /** \note Reaching the \#else means that we are finished with the
        *       then-body.
        */
      void
      executeElseStatement(ModuleInfo & moduleInfo);


      /// Parse the loop statement and return the corresponding loop context
      /** \pre scannedLine.m_LineContent.front() == \c \#loop
        * \param scannedLine The line starting with \#loop
        * \param nextLine Instruction pointer to the line following the \#loop
        *                 statement
        */
      LoopContext
      parseLoopStatement( Scanner::ScannedLine scannedLine,
                          InstructionPointer nextLine ) const;

      /// Execute the loop statement
      /** This method basically just pushes the loopContext onto the stack.
        */
      void executeLoopStatement( ModuleInfo& moduleInfo,
                            LoopContext loopContext);

      /// Execute \#eol statement
      /** \note There is no parseEndOfLoopStatement, since there is nothing
        *       to parse after the \#eol
        */
      void
      executeEndOfLoop ( ModuleInfo &moduleInfo );


      /// Parse an IRQ MASK statement
      /** \pre scannedLine.m_LineContent.front() == \c MASK
        * \return A list of InterruptIDs for interrupts that have to be
        *         enabled or disabled.
        */
      std::list< std::pair<InterruptID, bool> >
      parseIrqMaskStatement(Scanner::ScannedLine &scannedLine) const;

      /// Enable/Disable the specified interrupts
      void setIrqMask(const std::list< std::pair<InterruptID, bool> > &interruptEnableList);

      /// Enable/Disable the specified interrupts for the module
      void setIrqMaskForModule(const std::list< ::std::pair<InterruptID, bool> > &interruptEnableList, IFS_ModuleBase* module);

      /// Enable/disable the given interrupt
      void setInterruptEnableFlag(InterruptID interruptToBeEnabled,
                                  bool enable);


      /// Execute a ModuleCommand
      /** Execute the given user-defined module command.
        * \param scannedLine
        *   The line holding the command to execute (without the module's name).
        * \param pRequestingModule
        *   The module requesting the command.
        * \param instructionPointer
        *   Points to the line with the actually executed command.
        * \param moduleInfo
        *  The context of the current module
        * \return The kind of ModuleCommand that was executed.
        * \note There is no parseModuleCommand(...) method, because there is
        *       (nearly) nothing to parse.
        */
      ModuleCommandKind
      executeModuleCommand(
        Scanner::ScannedLine scannedLine,
        IFS_ModuleBase* pRequestingModule,
        InstructionPointer instructionPointer,
        const ModuleInfo& moduleInfo );


      /// Parse an IFS/ALL SYNC statement
      /** \pre scannedLine.m_LineContent.front() == \c SYNC
        * \return The list of modules to be synced
        * \note This parse method is non-static, because it does checking
        *       about the existence of the specified modules and therefore needs
        *       access to the controler's state.
        */
      std::list<const IFS_ModuleBase *>
      parseSyncStatement(Scanner::ScannedLine &scannedLine) const;

      /// Perform a \c SYNC for a list of modules
      /**
        * \param pRequestingModule The module that reached the SYNC
        * \param instructionPointer The instruction pointer for the SYNC
        * \param modulesToBeSynced  List of modules that have to be SYNCED.
        *
        * \note If the pRequestingModule is not element of modulesToBeSynced
        *       nothing is done.
        */
      void executeSync(
        const IFS_ModuleBase *pRequestingModule,
        InstructionPointer instructionPointer,
        const std::list<const IFS_ModuleBase*> &modulesToBeSynced);


      /// Parsed version of the WAIT statement
      struct WaitStatement
      {
        /// Kind of this wait statement
        enum Kind { simtime, deltime, cycles } m_Kind;

        /// Position of this WAIT statement
        InstructionPointer  m_InstructionPointer;

        /// The simulation time we are waiting for (valid only if m_Kind == simtime)
        sc_time       m_SimulationTimeEnd;
        /// The duration we want to wait (valid only if m_Kind == deltime)
        sc_time       m_WaitingDuration;
        /// The number of cycles we want to wait (valid only if m_Kind == cycles)
        unsigned int  m_NumberOfCycles;
      };


      /// Parse a WAIT statement
      /** \pre scannedLine.m_LineContent.front() == \c WAIT
        */
      WaitStatement parseWaitStatement( Scanner::ScannedLine scannedLine,
                                        InstructionPointer instructionPointer ) const;

      /// Execute the waitStatement for the given module
      void
      executeWaitStatement( IFS_ModuleBase *pRequestingModule,
                            WaitStatement waitStatement);

      /// \brief
      /// Continue an interrupted WAIT instruction (or issue a warning if the
      /// time is already exceeded).
      /// \note We continue waiting and do not proceed with the execution.
      /** This will suspend the command loop. Note this wait may be interrupted,
        * again.
        */
      void continueInterruptedWait(ModuleInfo &moduleInfo,
                                   const std::string &moduleName);


      /// The parsed version of a \#switch statement
      struct SwitchStatement
      {
        /// The condition following the \#switch
        std::string           m_Condition;

        /// The parsed version of a \#case statement
        struct CaseLabel
        {
          /// The label
          std::string         m_Value;
          /// The line following the \#case statement
          InstructionPointer  m_BlockBegin;
        };

        /// All \#case lables belonging to this \#switch
        std::list<CaseLabel>  m_CaseLabels;
        /// The line containing the \#endswitch
        InstructionPointer    m_EndOfSwitch;
      };

      /// Parse \#switch statement
      /** \pre scannedLine.front() == \#switch
        */
      SwitchStatement
      parseSwitchStatement(Scanner::ScannedLine const& scannedLine,ModuleInfo & moduleInfo,
                           InstructionPointer nextLine) const;

      /// Execute \#switch statement, i.e., jump to the corresponding \#case
      void
      executeSwitchStatement(ModuleInfo & moduleInfo,
                             SwitchStatement const& switchStatement);

      /// Execute \#case XYZ, i.e., jump to the corresponding \#endswitch
      void
      executeCaseStatement(ModuleInfo & moduleInfo);

      /// Execute \#case XYZ, i.e., jump to the corresponding \#endswitch
      void
      executeCriticalStatement(ModuleInfo & moduleInfo,
      const IFS_ModuleBase *pRequestingModule,
      Scanner::ScannedLine scannedLine,
      InstructionPointer nextLine);


      /// Parse EXPRESSION_ACCURACY command
      /** \pre scannedLine.m_LineContent.front() == \c EXPRESSION_ACCURACY
        */
      sc_fxtype_params
      parseExpressionAccuracy( Scanner::ScannedLine scannedLine );


      /// Return from an interrupt.
      /** Let a module return from executing commands in an interrupt-handler,
        * and possibly continue a WAIT or a SYNC.
        * \note To be called on \#end.
        * \param moduleInfo
        *    A structure holding data about the module returning from the
        *    interrupt.
        * \param moduleName
        *   The name of the module that executes the \#end (only for messages).
        * \param currentInstruction
        *    A pointer to the line containing the \#end
        * \param scannedLine
        *    The line with all the information about its source.
        */
      void returnFromISR( ModuleInfo &moduleInfo,
                          const std::string &moduleName,
                          InstructionPointer currentInstruction,
                          const Scanner::ScannedLine &scannedLine ) const;


      /// Indicates whether the specified list of Tokens represents a range.
      /** A range is combined of a ':' between two integers.
        * \return
        *   True, if they are a range, else false.
        * \note The values are not checked for being really integers.
        */
      bool
      isARange( const std::list<Scanner::Token>& checkedList ) const;

      /** Finds variables for the static script checking
        * \param variables
        * A container that contains the global variables
        * \param arrayVariables
        * A container that contains the global variables with more than one value (The arrays)
        */
      void
      findVariablesForTheStaticScriptChecking(
                                 std::map<std::string, std::string> &variables,
                                 std::map<std::string, std::list<std::string> > &arrayVariables) const;

      /** Checks an isr command
        * \param scannedLine
        * Represents the informations about the command line.
        * \param lineNumber
        *  Represents the current line number in the script.
        * \param variableWasFound
        * Stores the information whether a variable was found or not
        * \return
        *  True, if an error was found, else false.
        */
      bool
      checkISRCommand(IFS::Scanner::ScannedLine& scannedLine, const std::string& lineNumber, bool variableWasFound) const;

      /** Checks a loop command
        * \param scannedLine
        * Represents the informations about the command line.
        * \param lineNumber
        *  Represents the current line number in the script.
        * \param lc
        * Contains the current loop contexts
        * \param variableWasFound
        * Stores the information whether a variable was found or not
        * \return
        *  True, if an error was found, else false.
        */
      bool
      checkLoopCommand(IFS::Scanner::ScannedLine& scannedLine, const std::string& lineNumber, std::list<LoopContext> &lc, bool variableWasFound) const;

      /** Checks an if command
        * \param scannedLine
        * Represents the informations about the command line.
        * \param lineNumber
        *   Represents the current line number in the script.
        * \param variableWasFound
        * Stores the information whether a variable was found or not
        * \return
        *  True, if an error was found, else false.
        */
      bool
      checkIfCommand(IFS::Scanner::ScannedLine& scannedLine, const std::string& lineNumber, bool variableWasFound) const;

      /** Checks a switch command
        * \param scannedLine
        * Represents the informations about the command line.
        * \param lineNumber
        *  Represents the current line number in the script.
        * \param variableWasFound
        * Stores the information whether a variable was found or not
        * \return
        *  True, if an error was found, else false.
        */
      bool
      checkSwitchCommand(IFS::Scanner::ScannedLine& scannedLine, const std::string& lineNumber, bool variableWasFound) const;

      /** Checks a critical command
        * \param scannedLine
        * Represents the informations about the command line.
        * \param lineNumber
        *  Represents the current line number in the script.
        * \param variableWasFound
        * Stores the information whether a variable was found or not
        * \return
        *  True, if an error was found, else false.
        */
      bool
      checkCriticalCommand(IFS::Scanner::ScannedLine& scannedLine, const std::string& lineNumber, bool variableWasFound) const;

      /** Checks a case command
        * \param scannedLine
        * Represents the informations about the command line.
        * \param lineNumber
        *  Represents the current line number in the script.
        * \param variableWasFound
        * Stores the information whether a variable was found or not
        * \return
        *  True, if an error was found, else false.
        */
      bool
      checkCaseCommand(IFS::Scanner::ScannedLine& scannedLine, const std::string& lineNumber, bool variableWasFound) const;

      /** Checks whether a line consists of one token or not
        * \param scannedLine
        * Represents the informations about the command line.
        * \param lineNumber
        *  Represents the current line number in the script.
        * \param variableWasFound
        * Stores the information whether a variable was found or not
        * \return
        *  True, if an error was found, else false.
        */
      bool
      checkIfTheLineConsistsOfOneToken(IFS::Scanner::ScannedLine& scannedLine, const std::string& lineNumber, bool variableWasFound) const;

      /** Checks a control command
        * \param scannedLine
        * Represents the informations about the command line.
        * \param lineNumber
        *  Represents the current line number in the script.
        * \param variableWasFound
        * Stores the information whether a variable was found or not
        * \return
        *  True, if an error was found, else false.
        */
      bool
      checkControlCommand(IFS::Scanner::ScannedLine& scannedLine, const std::string& lineNumber, bool variableWasFound) const;

      /** Checks a module command
        * \param scannedline
        * Represents the informations about the command line.
        * \param lineNumber
        *  Represents the current line number in the script.
        * \param variableWasFound
        * Stores the information whether a variable was found or not
        * \return
        *  True, if an error was found, else false.
        */
      bool
      checkModuleCommand(IFS::Scanner::ScannedLine& scannedline, const std::string& lineNumber, bool variableWasFound) const;

      /** Checks whether a module is registered, or not
        * \param moduleName
        * The module name which ought to be checked
        *  True, if the module is registered, else false.
        */
      bool
      isModuleRegistered(const std::string &moduleName) const;

      ///execute TITLE Command
      void
      executeTitleStatement(IFS::Scanner::ScannedLine& scannedLine, const IFS::IFS_ModuleBase* pRequestingModule);

      /** Checks whether a command is registered, or not
        * \param moduleName
        * Indicates where the command is registered.
        * \param commandName
        * The command name which ought to be checked.
        * \return
        * True, if the command is registered, else false.
        */
      bool
      isCommandRegistered(const std::string &moduleName, const std::string &commandName) const;

      /** Checks whether an IFS command is a unique control command (PRINT and WAIT can be a module command)
       * \param scannedLine
       *  The line to be checked
       * */
      bool isControlCommand(IFS::Scanner::ScannedLine& scannedLine) const;
      /// @}

      /// randomise helper method
      void
      displayRandomParameters(
        const std::list< std::list<IFS::Scanner::Token> >& parameterList,
        const std::list< unsigned int >& probabilityList,
        const std::list< bool >&         probabilitySpecificationList,
        const std::string &fileName,
        unsigned int lineNumber ) const;
  };

} // Impl
} // IFS

#endif

