/*
 *  (C) Robert Bosch GmbH 2008. All rights reserved, also regarding any
 *  disposal, exploitation, reproduction, editing, distribution, as
 *  well as in the event of application for industrial property rights.
 *
 *  Contact: AE/NE1-IFS <vAENE1.IFS@de.bosch.com>
 *
 * Disclaimer:
 * This Software is licensed under the following conditions. You may 
 * not use this file except in compliance with the following conditions:
 *
 * Robert Bosch GmbH hereby grants to licensee the non-exclusive and 
 * non-transferable right to use the software and accompanying 
 * documentation for the sole purpose of testing designs delivered by 
 * Robert Bosch GmbH to licensee.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND WITHOUT ANY WARRANTY,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE, TITLE AND 
 * NONINFRINGEMENT. IN NO EVENT SHALL ROBERT BOSCH GMBH BE LIABLE FOR
 * ANY DIRECT OR INDIRECT DAMAGES, INJURY OR OTHER LIABILITY, ARISING
 * OUT OR IN CONNECTION WITH THE SOFTWARE. ROBERT BOSCH GMBH SHALL
 * NOT BE LIABLE FOR THE ACCURACY OF THE DESIGNS TESTED WITH THE AID
 * OF THE SOFTWARE.
*/

#ifndef IFS_EXPRESSION_INCLUDED
#define IFS_EXPRESSION_INCLUDED

#define SC_INCLUDE_FX
#include <systemc.h>
#include <string>


#include "IFS_ErrorMessage.h"
#include "IFS_Scanner.h"
#include "IFS_OutputHandler.h"

namespace IFS
{
  /// Contains the expression evaluation functions
  namespace Expression
  {
    /// All arithmetic will be based on sc_ufix
    /** \note When changing this, you should check
      *  IFS_Expression.c++/struct ifsNumberParser::operator () (...)
      *
      */
    typedef sc_ufix ArithmeticExpressionType;

    /// Check whether s complies to the syntax of the given number type
    bool
    isIFSNumber(std::string const& s);

    /// Try to parse and evaluate the given string, which must be an IFS-number
    ArithmeticExpressionType
    evalIFSNumber(std::string const& s,
                   sc_fxtype_params evaluationAccuracy ) throw( IFS_ErrorMessage );

    /// Try to parse and evaluate the given string, which must be an IFS floating point number
    double
    evalIFSFloatingPointNumber(std::string const& s) throw( IFS_ErrorMessage );

    /// Replace occurrences of $( ... ) expressions by their corresponding value
    void
    substituteExpressions( Scanner::ScannedLine &scannedLine,
                           sc_fxtype_params evaluationAccuracy );

    //checks whether the shift-operation is ok
    void checkShift(ArithmeticExpressionType str);
  }
}

// -----------------------------------------------------------------------------

// optimised implementation
inline
bool
IFS::Expression::isIFSNumber(std::string const& s)
{
  std::string::value_type const* p = s.c_str();

  enum Base { hex, dez, bin } base=dez;

  switch (*p)
  {
    case '\0' :
      // empty string
      return false;
    case 'b' :
    case 'B' :
      base = bin;
      break;
    case 'h' :
    case 'H' :
      base = hex;
      break;
    case 'd' :
    case 'D' :
      break;
    default:
      return false;
  }

  ++p;

  if (*p != '#')
  {
    return false;
  }

  ++p;

  // there must be at least one character after the prefix
  if (*p == '\0')
  {
    return false;
  }

  switch (base)
  {
    case dez:
      while(*p)
      {
        if (*p < '0' || *p > '9')
        {
          return false;
        }
        ++p;
      }
      return true;
    case bin:
      while(*p)
      {
        if (*p < '0' || *p > '1')
        {
          return false;
        }
        ++p;
      }
      return true;
    case hex:
      while(*p)
      {
        if (!( (*p >= '0' && *p <= '9') ||
               (*p >= 'a' && *p <= 'f') ||
               (*p >= 'A' && *p <= 'F')
             )
           )
        {
          return false;
        }
        ++p;
      }
      return true;

    default:
      return false;
  }

  return false;
}

#endif

