/*
 *  (C) Robert Bosch GmbH 2008. All rights reserved, also regarding any
 *  disposal, exploitation, reproduction, editing, distribution, as
 *  well as in the event of application for industrial property rights.
 *
 *  Contact: AE/NE1-IFS <vAENE1.IFS@de.bosch.com>
 *
 * Disclaimer:
 * This Software is licensed under the following conditions. You may 
 * not use this file except in compliance with the following conditions:
 *
 * Robert Bosch GmbH hereby grants to licensee the non-exclusive and 
 * non-transferable right to use the software and accompanying 
 * documentation for the sole purpose of testing designs delivered by 
 * Robert Bosch GmbH to licensee.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND WITHOUT ANY WARRANTY,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE, TITLE AND 
 * NONINFRINGEMENT. IN NO EVENT SHALL ROBERT BOSCH GMBH BE LIABLE FOR
 * ANY DIRECT OR INDIRECT DAMAGES, INJURY OR OTHER LIABILITY, ARISING
 * OUT OR IN CONNECTION WITH THE SOFTWARE. ROBERT BOSCH GMBH SHALL
 * NOT BE LIABLE FOR THE ACCURACY OF THE DESIGNS TESTED WITH THE AID
 * OF THE SOFTWARE.
*/

#ifndef IFS_MACROS_H
#define IFS_MACROS_H

#include <systemc.h>

// Modelling-related macros
//==============================================================================

#define IFS_MODULE(moduleName) struct moduleName : public IFS::IFS_ModuleBase

#define IFS_LEGACY_MODULE( moduleName ) \
  class moduleName : public IFS::IFS_WrapperModule

//------------------------------------------------------------------------------

#define IFS_REGISTER_COMMAND( moduleType, commandString, commandMethod) \
  registerCommand( \
    commandString, \
    new IFS::IFS_CommandPointer<moduleType>( \
      this, \
      &moduleType:: commandMethod) );

//------------------------------------------------------------------------------

#define IFS_REGISTER_VALUE( nameOfTheValue, pointerOfTheValue) \
  registerValue( \
    nameOfTheValue, \
    pointerOfTheValue);

//------------------------------------------------------------------------------

#define IFS_REGISTER_ADDRESS( label, address) \
  IFS::registerAddress(label, address);

//------------------------------------------------------------------------------

#define IFS_REGISTER_CLOCK_EVENT( clockEvent ) \
  registerClockEvent( clockEvent );

//------------------------------------------------------------------------------

#define IFS_SET_CHECKVALUE( value ) \
  setCheckValue( value );

//------------------------------------------------------------------------------

#define IFS_REQUEST_INTERRUPT( interruptID ) \
  interruptOccurred( interruptID );

//------------------------------------------------------------------------------

#define IFS_OUT(debugLevel) ifsout(IFS::IFS_Message::debugLevel)

//------------------------------------------------------------------------------


#if (SYSTEMC_VERSION > 20011005) && (SYSTEMC_VERSION <= 20020405)
  // its SystemC 2.0.1
  #define IFS_GET_PROCESS_NAME_AS_STRING ( std::string( sc_get_curr_simcontext() \
                                        ->get_curr_proc_info() \
                                        ->process_handle->name() ) )
  #define IFS_GET_MODULE_NAME_AS_STRING ( std::string( sc_get_curr_simcontext() \
					->get_curr_proc_info() \
			  		->process_handle->module->name() ) )
#elif SYSTEMC_VERSION == 0
  // its NCSim
  #define IFS_GET_PROCESS_NAME_AS_STRING ( std::string( sc_get_current_process_handle() \
                                        .name() ) )
  // TODO: This will fail with more complex strctures, would need a loop and
  // dynamic casts to check for the proper parent. But it would also work
  // for older SystemC versions.
  #define IFS_GET_MODULE_NAME_AS_STRING ( std::string( sc_get_current_process_handle() \
                                          .get_parent_object()->name() ) )
#elif SYSTEMC_VERSION <= 20050714
  // its SystemC 2.1
  #define IFS_GET_PROCESS_NAME_AS_STRING ( std::string( sc_get_curr_process_handle() \
                                        ->name() ) )
  #define IFS_GET_MODULE_NAME_AS_STRING ( std::string( sc_get_curr_process_handle() \
					->host->name() ) )
#elif SYSTEMC_VERSION <= 20070314
  // its SystemC 2.2
  #define IFS_GET_PROCESS_NAME_AS_STRING ( std::string( sc_get_current_process_handle() \
                                        .name() ) )
  // TODO: This will fail with more complex strctures, would need a loop and
  // dynamic casts to check for the proper parent. But it would also work
  // for older SystemC versions.
  #define IFS_GET_MODULE_NAME_AS_STRING ( std::string( sc_get_current_process_handle() \
                                          .get_parent_object()->name() ) )
#else
  // Unknown version. Try the implementation for SystemC 2.1
  #define IFS_GET_PROCESS_NAME_AS_STRING ( std::string( sc_get_curr_process_handle() \
                                        ->name() ) )
  #define IFS_GET_MODULE_NAME_AS_STRING ( std::string( sc_get_curr_simcontext() \
                                          ->get_curr_proc_info()->process_handle \
                                          ->host->name() ) )
#endif

// Error-related macros
//==============================================================================

#define INTERNAL_ERROR_MESSAGE( messageString ) \
  IFS_ErrorMessage( \
    messageString, \
    IFS_Message::IFS, \
    IFS_Message::formatFileMessageSourceInfo( \
      std::string(__FILE__), \
      __LINE__) +  \
    std::string(", in function: ") + \
    std::string(__FUNCTION__) \
  )

//------------------------------------------------------------------------------

#define IFS_ERROR_MESSAGE( messageString ) \
  IFS_ErrorMessage( \
    messageString, \
    IFS_Message::IFS \
  )

//------------------------------------------------------------------------------
#define IFS_MODULE_ERROR( messageString ) \
  IFS::IFS_ErrorMessage( \
    messageString, \
    IFS::IFS_Message::Module, \
    name())

// Coverage-related macros
//==============================================================================

#define IFS_COVERAGE(count,last_named_arg,...) \
	  IFS::m_CoverageBase.coverage_status(count,last_named_arg,__VA_ARGS__);

//------------------------------------------------------------------------------

// macro to check coverage status
#define IFS_COVERAGE_STATUS() \
	  IFS::m_CoverageBase.coverage_status();

//------------------------------------------------------------------------------

// registers coverage signal
#define IFS_REGISTER_COVERAGE_SIGNAL(signal_name,signal) \
		try \
		{ \
	  	IFS::m_CoverageBase.registerCoverageSignal(signal_name,&signal);\
		} \
	  catch(IFS::IFS_ErrorMessage& caughtError )\
    {\
      g_IfsOut << caughtError;\
      g_IfsOut << IFS_Message(\
        "Skipping Registration.",\
        IFS_Message::Module,\
        "",\
        IFS_Message::Warning);\
    }

// Constraint-related macros
//==============================================================================

//macro to register constraint with name and default value
#define IFS_REGISTER_CONSTRAINT(ifs_ptr, ifs_ptr_name, default_value) \
	  this->registerConstraint(&ifs_ptr, ifs_ptr_name,(typeof(ifs_ptr.read()))default_value);

//------------------------------------------------------------------------------

#define IFS_ADAPT_CONSTRAINT(constraint_class, constraint) \
		if(constraint.get_name() == constraint_name) \
		{ constraint_class->adapt_constraint(constraint, parameters);} \

//==============================================================================
#endif

