/*
 *  (C) Robert Bosch GmbH 2008. All rights reserved, also regarding any
 *  disposal, exploitation, reproduction, editing, distribution, as
 *  well as in the event of application for industrial property rights.
 *
 *  Contact: AE/NE1-IFS <vAENE1.IFS@de.bosch.com>
 *
 * Disclaimer:
 * This Software is licensed under the following conditions. You may 
 * not use this file except in compliance with the following conditions:
 *
 * Robert Bosch GmbH hereby grants to licensee the non-exclusive and 
 * non-transferable right to use the software and accompanying 
 * documentation for the sole purpose of testing designs delivered by 
 * Robert Bosch GmbH to licensee.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND WITHOUT ANY WARRANTY,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE, TITLE AND 
 * NONINFRINGEMENT. IN NO EVENT SHALL ROBERT BOSCH GMBH BE LIABLE FOR
 * ANY DIRECT OR INDIRECT DAMAGES, INJURY OR OTHER LIABILITY, ARISING
 * OUT OR IN CONNECTION WITH THE SOFTWARE. ROBERT BOSCH GMBH SHALL
 * NOT BE LIABLE FOR THE ACCURACY OF THE DESIGNS TESTED WITH THE AID
 * OF THE SOFTWARE.
*/

#ifndef IFS_MESSAGE_H
#define IFS_MESSAGE_H

#include <string>

namespace sc_core {
  class sc_time;
}
using sc_core::sc_time;

namespace IFS
{
  /// Class for storing all kinds of messages (to be displayed to the user)
  class IFS_Message
  {
    public:
      /// Classification of the severity of a message
      /** \note The sequence of the elements has a meaning and must not
        *       be changed.
        */
      enum Severity { Debug, Info, Warning, Error };

      /// \brief Source of the error, i.e. where the error came from
      enum Source { IFS, Module, Testbench, Script, Utils };

      /// Constructor
      /**
        * \param message  Some relevant text
        * \param source   The source of this message (or the expected source
        *                 of an error in case of an error message)
        * \param additionalSourceInfo More details about the source (e.g.
        *                             file name, line number, ...)
        * \param severity Either Info, Warning or Error
        *
        */
      IFS_Message(const std::string & message,
                  Source source,
                  const std::string & additionalSourceInfo = "",
                  Severity severity = Error);

      /// Copy constructur (needed for proper throwing)
      IFS_Message(const IFS_Message &sourceMessage);

      /// Assignment
      IFS_Message &
      operator = (const IFS_Message &sourceMessage);

      /// Destructor
      virtual ~IFS_Message();

      /// Get severity of this message (as simple as it sounds)
      Severity getSeverity() const;
      /// Get the message
      const std::string& getMessage() const;
      /// Get the source
      Source getSource() const;
      /// Get the additional source info
      const std::string& getSourceInfo() const;

      /// \brief
      /// Return a (nicely formatted) string containing message, source,
      /// severity, simulation time, etc.
      /** \note Maybe add some kind of format string to control the
        *       appearance of the returned string.
        */
      virtual std::string getFormattedMessage() const;

      /// \brief
      /// Generate a string containing file name and line number in a consistent
      /// way. Purpose: to be used in additionalSourceInfo for messages/errors
      static std::string
      formatFileMessageSourceInfo(const std::string& fileName,
                                  unsigned int lineNumber);

      /// \brief
      /// Same as above, but with more specific info about where in the line
      /// the problem is.
      static std::string
      formatFileMessageSourceInfo(const std::string& fileName,
                                  unsigned int      lineNumber,
                                  std::string::size_type columnBegin,
                                  std::string::size_type columnEnd );

      /// \brief
      /// Same as above, but less specific (when there is no meaningful line to
      /// reference, e.g., when something is missing).
      static const std::string&
      formatFileMessageSourceInfo(const std::string& fileName);

    private:

      /// Message (provided by the constructor)
      std::string  m_Message;
      /// Source (provided by the constructor)
      Source  m_Source;
      /// Additional source info (provided by the constructor)
      std::string  m_SourceInfo;
      /// Severity (provided by the constructor)
      Severity m_Severity;
      /// Time of creation (may be different from the time when it is displayed)
      sc_time *m_pCreationTime;
  };
}

#endif

