/*
 *  (C) Robert Bosch GmbH 2008. All rights reserved, also regarding any
 *  disposal, exploitation, reproduction, editing, distribution, as
 *  well as in the event of application for industrial property rights.
 *
 *  Contact: AE/NE1-IFS <vAENE1.IFS@de.bosch.com>
 *
 * Disclaimer:
 * This Software is licensed under the following conditions. You may 
 * not use this file except in compliance with the following conditions:
 *
 * Robert Bosch GmbH hereby grants to licensee the non-exclusive and 
 * non-transferable right to use the software and accompanying 
 * documentation for the sole purpose of testing designs delivered by 
 * Robert Bosch GmbH to licensee.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND WITHOUT ANY WARRANTY,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE, TITLE AND 
 * NONINFRINGEMENT. IN NO EVENT SHALL ROBERT BOSCH GMBH BE LIABLE FOR
 * ANY DIRECT OR INDIRECT DAMAGES, INJURY OR OTHER LIABILITY, ARISING
 * OUT OR IN CONNECTION WITH THE SOFTWARE. ROBERT BOSCH GMBH SHALL
 * NOT BE LIABLE FOR THE ACCURACY OF THE DESIGNS TESTED WITH THE AID
 * OF THE SOFTWARE.
*/

#ifndef IFS_MODULE_BASE_H
#define IFS_MODULE_BASE_H

#include "IFS_CommandPointerBase.h"
#include "IFS_Message.h"

#include <systemc.h>

#include <list>
#include <map>
#include <string>


/// This namespace holds all IFS-related classes and functions.
namespace IFS
{

  // Prototypical class-declarations
  class IFS_Control;
  namespace Impl
  {
    class IFS_Control;
  }

  /// \brief Identifier for an interrupt (and its corresponding handler)
  /** \note This is defined in IFS_ModuleBase.h, too. However, spending
    *       an extra header file for this seems to be overkill.
    */
  typedef unsigned int InterruptID;

  /// Base class for user-defined IFS modules
  /** This class acts as interface between the user's IFS module and the
    * controller. Furthermore it provides some predefined commands.
    */
  // The 'struct'-version (instead of SC_MODULE) of the module-declaration is
  // used here in order to make life easier for doxygen.
  class IFS_ModuleBase: public sc_module
  {
    public:
      SC_HAS_PROCESS(IFS_ModuleBase);

      /// Constructor
      IFS_ModuleBase( sc_module_name moduleName, bool legacyMode = false );

      /// Destructor
      virtual ~IFS_ModuleBase() {}

      /// Callback for user-defined initialisation code
      /** This function is always called before the command-loop of a module is
        * started. So a user can overload it with a function carrying his/her own
        * initialisations.
        * \note The default-function is just a dummy doing nothing.
        */
      virtual void initialiseModule() {}

      /// Callback for user-defined finalisation code
      /** This function is always called when a module is shutting down
        * So a user can overload it with a function carrying his/her own
        * finalisations.
        * \note The default-function is just a dummy doing nothing.
        */
      virtual void finaliseModule() {}

      /// This methods returns the currently executed line number.
      /// \return The line number.
      unsigned int getLineNumber() const;

      /// This methods returns the name of the currently executed command file.
      /// Due to include statements the name can change during simulation.
      /// \return The filename.
      const std::string& getFilename() const;

    protected:

      /// \brief
      /// Pass coverage information to a specific command file variable
      ///
      /// \param var_name = name of command file variable and coverage signal
      /// \param var_value = assigned value
      void passCoverageInfoToCmdFileVariable(const std::string& var_name, const std::string& var_value);

      /// \brief
      /// This method maps the values of variables to the corresponding registered string variables.
      /// This method does nothing by default. It has to be implemented in the child class.
      virtual void updateGetValueVariables();

      /** Assigns the information about the script file and command line, which
       * are currently executed, to the internal signals scriptLine and scriptFile.
       *  \param lineContent Contains the command line
       *  \param filename Contains the filename of the script. May change during simulation due to include statements
       *  \param lineNumber Contains the line number
       */
      void setSignals(const std::string& lineContent, const std::string& filename, const unsigned int lineNumber);

      /// \defgroup Methods_for_the_user Methods to be used by the user-defined IFS modules (mostly via macros)
      /// @{

      /// Registers a user-defined command.
      /** Registers a user-defined command. This should be done before the simulation
        * starts (e.g. in a constructor or the main()-function).
        * \param newCommandString
        *   Name of the command in the script
        * \param pNewCommandPointer
        *   Pointer to the method that will be called when the command is
        *   invoked.
        */
      void registerCommand(const std::string& newCommandString,
              IFS::IFS_CommandPointerBase *const pNewCommandPointer );

      //Registers a user-defined value
      /** Registers a user-defined value. This should be done before the
        * simulation starts (e.g. in a constructor or the main()-function).
        * \param newValueString
        *   Name of the value in the script
        * \param newValuePointer
        *   Pointer to the value
        */
      void registerValue(const std::string& newValueString, std::string *const newValuePointer);

      /// \brief
      ///   Register the event to be used in WAIT cycles (use .pos() in order
      ///   to obtain the event finder from a port)
      ///
      void registerClockEvent( const sc_event_finder& clockEventFinder );

      /// Notify the controller about interrupts that occured simultaneously.
      void interruptsOccurred( std::list<InterruptID> occurredInterrupts);

      /// Notify the controller about the specified interrupt
      void interruptOccurred(InterruptID occurredInterrupt);

      /// legacy print command
      void print(std::list<std::string> parameterList);
      void print(const std::string& parameterString);

      /// \brief
      /// This method stores a string which is referred to by the CHECK-command
      /// in command-scripts.
      /** \note: See getCheckValue() and isCheckValueSet().
        */
      void setCheckValue( const std::string& checkValue );

      /// The IFS output channels as standard stream
      std::ostream &ifsout(IFS_Message::Severity severity=IFS::IFS_Message::Info) const;

      /// @}

      /// Waits for the specified number of cycles.
      /**
        * The waiting is not interruptable.
        * \note This wait variant can be handled locally by the module, hence
        *       it is implemented here and not in the controller (as opposed
        *       to the other two WAIT variants).
        * \note
        *   See also registerClockEvent().
        */
      virtual void waitCycles( unsigned int numberOfCycles );

      /// \brief
      /// This method allows to define the way modules wait on external command requests.
      /// Usually used for LegacyModules only.
      virtual void waitExternalCommandRequest();

  public:
      /// The number of characters of the script file name
      #define SCRIPTFILE_NUM_CHARS 50

      /// The number of characters of the script line
      #define SCRIPTLINE_NUM_CHARS 100

      // Stores the information about the script line, which is currently executed.
      sc_signal< sc_bv<8*SCRIPTLINE_NUM_CHARS> > scriptLine;

      // Stores the information about the script file,which is currently executed.
      sc_signal< sc_bv<8*SCRIPTFILE_NUM_CHARS> > scriptFile;

      // stores the information from the command TITLE
      #define TITLE_SIZE 100
      sc_signal< sc_bv<8*TITLE_SIZE> > title0;
      sc_signal< sc_bv<8*TITLE_SIZE> > title1;
      sc_signal< sc_bv<8*TITLE_SIZE> > title2;
      sc_signal< sc_bv<8*TITLE_SIZE> > title3;
      sc_signal< sc_bv<8*TITLE_SIZE> > title4;

    private:

      //// stores the title-variables
      std::vector<sc_signal< sc_bv<8*TITLE_SIZE> >* > m_listOfTitles;

      /// \brief
      /// The command loop (SC_THREAD) calls the controller in order to execute
      /// commands from the script.
      void commandLoop();

      /** \defgroup Methods_for_the_controller Methods to be used by the controller for accessing user-defined IFS modules
        * @{
        */

      /// \brief
      /// The controller is a module's friend. This way we prevent a user from
      /// calling some of the module's methods.
      friend class ::IFS::Impl::IFS_Control;

      /// Set the controller (done during registration)
      void setController( IFS::IFS_Control* pNewController );

      /// \brief
      /// Get the available commands from this module (called
      /// during registration)
      const std::map< std::string, IFS::IFS_CommandPointerBase* > &
        getCommandMapping() const;

      /// The event, which is waited for in waitCycles().
      const sc_event & getClockEvent() const;

      /// Resets the state of the check-value to being unset.
      /** \note: See isCheckValueSet().
        */
      void resetCheckValue();

      /// Indicates whether a check-value has been set or not.
      /** The initial state is unset. This changes, if a value is stored with
        * setCheckValue(). Afterwards it is set until a call of
        * resetCheckValue() occurred, when it becomes unset, again.
        * \return
        *   'true', if a value has been set, else 'false'.
        */
      bool isCheckValueSet() const;

      /// Returns the string representing the to-be-checked-value.
      /** Returns the string representing the to-be-checked-value, unless it is
        * in an unset state, in which case an exception (IFS_ErrorMessage) is
        * thrown.
        * \return The value set last with setCheckValue().
        * \note: See resetCheckValue(), setCheckValue() and isCheckValueSet().
        */
      std::string getCheckValue() const;

      /// Returns the string representation of a value which is registered in the value map.
      /** Returns the string representation of a value which is registered in the value map.
        * \param nameValueString
        * Indicates which value has to be returned.
        * \return
        * If an value exists the value will be returned, otherwise an error will occur.
        */
      std::string getValue(const std::string& nameValueString) const;

      /// Indicates if a module is working in legacy-mode.
      /** \return
        *   'true', if the module is operating in legacy-mode, else 'false'.
        */
      bool isInLegacyMode() const;

      /// \brief
      /// Required for legacy-mode: Transfer the specified
      /// string to the attached legacy-module of this module.
      /** \note This method is overriden by
        *       IFS_WrapperModule::transferLegacyCommand(const string& commandString)
        *       and must only be called for IFS_WrapperModule instances. The
        *       implementation of this class will throw an exception.
        */
      virtual void transferLegacyCommand( const std::string& commandString );

      /// @}

  protected:
      /// Pointer to the controller where this module is registered
      IFS::IFS_Control *m_pController;

  private:
      /// Map of commands
      std::map< std::string, IFS::IFS_CommandPointerBase* > m_CommandMapping;

       /// Map of values
      std::map< std::string, std::string* > m_ValueMapping;

      /// Needed to defer accessing the clock event (defer until the port is bound)
      const sc_event_finder *m_pClockEventFinder;

      /// \brief
      /// Required for legacy-mode:
      /// Indicates, whether the module is in legacy-mode or not.
      bool m_LegacyMode;

      /// Indicates whether the check-value is set or unset
      bool m_CheckValueIsSet;

      /// Holds the value of the to-be-checked-value
      std::string m_CheckValue;

      /// Stores the information which interrupts are valid for this module
      /// The informations can be set by the command
      /// '\<moduleName> IRQ MASK \<IrqMaskSpecifier>'
      std::map<InterruptID, bool> m_InterruptEnableList;

      /// Holds the currently executed line number
      unsigned int m_LineNumber;

      /// Holds the name of the file which contains the currently executed command
      std::string m_Filename;
  };
}

#endif

