/*
 *  (C) Robert Bosch GmbH 2008. All rights reserved, also regarding any
 *  disposal, exploitation, reproduction, editing, distribution, as
 *  well as in the event of application for industrial property rights.
 *
 *  Contact: AE/NE1-IFS <vAENE1.IFS@de.bosch.com>
 *
 * Disclaimer:
 * This Software is licensed under the following conditions. You may 
 * not use this file except in compliance with the following conditions:
 *
 * Robert Bosch GmbH hereby grants to licensee the non-exclusive and 
 * non-transferable right to use the software and accompanying 
 * documentation for the sole purpose of testing designs delivered by 
 * Robert Bosch GmbH to licensee.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND WITHOUT ANY WARRANTY,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE, TITLE AND 
 * NONINFRINGEMENT. IN NO EVENT SHALL ROBERT BOSCH GMBH BE LIABLE FOR
 * ANY DIRECT OR INDIRECT DAMAGES, INJURY OR OTHER LIABILITY, ARISING
 * OUT OR IN CONNECTION WITH THE SOFTWARE. ROBERT BOSCH GMBH SHALL
 * NOT BE LIABLE FOR THE ACCURACY OF THE DESIGNS TESTED WITH THE AID
 * OF THE SOFTWARE.
*/

#ifndef IFS_OUTPUT_HANDLER_H
#define IFS_OUTPUT_HANDLER_H

#include <list>
#include <string>
#include <fstream>

#include "IFS_Message.h"

namespace IFS
{
  /// The OutputHandler controls/handles the output of IFS_Messages
  class IFS_OutputHandler
  {
    public:

      /// Default constructor
      IFS_OutputHandler();

      /// destructor
      ~IFS_OutputHandler();

      /// Return a string corresponding to the given severity
      static std::string getChannelPrefix( const IFS_Message::Severity& channel );

      /// Output an IFS_Message
      const IFS_OutputHandler & operator << (const IFS_Message &message);

      /// Specifiy up to which severity level messages should be displayed
      void setReportLevel(IFS_Message::Severity reportLevel);

      /// disables the output
      void disableTheOutput();

      /// enables the output
      void enableTheOutput();

      /// sets the m_WriteTraceLog variable
      void setWriteTraceLog(bool value);

      /// sets the m_FileName variable
      void setFileName(const std::string& name);

      /// Get current report level
      IFS_Message::Severity getReportLevel() const;

      /// Suppress the output of specific messages
      /** Suppress messages with a severity <= the parameter severity and with a
        * source equalling the parameter source.
        * \param severity
        *   Specifies the severity of the messages that shall be suppressed.
        * \param source
        *   Specifies the source of the messages that shall be suppressed.
        * \param message
        *   and
        * \param sourceInfo
        *   are optional arguments. If a message matches one or both arguments
        *   (when non-empty) then the message is suppressed.
        */
      void suppressMessages( IFS_Message::Severity severity,
                             IFS_Message::Source source,
                             const std::string &message = "",
                             const std::string &sourceInfo = "" );

      /// Reset suppressions and report-level to none and INFO.
      /** Enable the output of all messages (i.e. clear suppressions, set
        * report level to Info)
        */
      void enableAllMessages();

      /// Returns the number of errors-messages displayed so far.
      unsigned int getNumberOfDisplayedErrors() const;

      /// Returns the number of warning-messages displayed so far.
      unsigned int getNumberOfDisplayedWarnings() const;

      /// \brief Returns the output stream that belongs to the given severity
      ///        (including the channel prefix)
      std::ostream &getOutputStream(IFS_Message::Severity severity=IFS_Message::Info);

    private:
      /// Only messages with a severity >= m_ReportLevel are displayed
      IFS_Message::Severity   m_ReportLevel;

      /// stores the information wheter the output is enabled or not
      bool m_disableOutput;

      /// Helper struct to store the suppression info
      struct SuppressionEntry
      {
        /// Messages with a severity <= m_Severity will be suppressed
        const IFS_Message::Severity m_Severity;
        /// Messages with a source == m_Source will be suppressed
        const IFS_Message::Source   m_Source;
        /// If != "" messages with a content == m_Message will be suppressed
        const std::string           m_Message;
        /// If != "" messages with a source == m_SourceInfo
        const std::string           m_SourceInfo;
      };

      /// Accumulate all suppressions in this list
      std::list<SuppressionEntry>   m_SuppressionList;

      /// Holds the number of errors displayed so far.
      unsigned int m_ErrorCounter;

      /// Holds the number of warnings displayed so far.
      unsigned int m_WarningCounter;

      /// the output stream for the trace file
      std::fstream m_OutputFile;

      /// determines whether an output file has to be written or not
      bool m_WriteTraceLog;

      /// contains the file name of the trace file
      std::string m_FileName;

      /// stores the information whether the file is opened the first time
      bool m_WriteFirstTime;
  };

  /// global instance of the output handler
  extern IFS_OutputHandler g_IfsOut;

  /// \brief A global function to obtain an output stream outside a module (e.g. the
  ///        testbench)
  std::ostream &ifsout(IFS_Message::Severity severity=IFS_Message::Info);

}

  /// \brief The using directive is needed in order to be able to write ifsout(...)
  ///        in the module (call IFS_ModuleBase::ifsout) and in the testbench
  ///        (call IFS::ifsout).
  using IFS::ifsout;

#endif

