/*
 *  (C) Robert Bosch GmbH 2008. All rights reserved, also regarding any
 *  disposal, exploitation, reproduction, editing, distribution, as
 *  well as in the event of application for industrial property rights.
 *
 *  Contact: AE/NE1-IFS <vAENE1.IFS@de.bosch.com>
 *
 * Disclaimer:
 * This Software is licensed under the following conditions. You may 
 * not use this file except in compliance with the following conditions:
 *
 * Robert Bosch GmbH hereby grants to licensee the non-exclusive and 
 * non-transferable right to use the software and accompanying 
 * documentation for the sole purpose of testing designs delivered by 
 * Robert Bosch GmbH to licensee.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND WITHOUT ANY WARRANTY,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE, TITLE AND 
 * NONINFRINGEMENT. IN NO EVENT SHALL ROBERT BOSCH GMBH BE LIABLE FOR
 * ANY DIRECT OR INDIRECT DAMAGES, INJURY OR OTHER LIABILITY, ARISING
 * OUT OR IN CONNECTION WITH THE SOFTWARE. ROBERT BOSCH GMBH SHALL
 * NOT BE LIABLE FOR THE ACCURACY OF THE DESIGNS TESTED WITH THE AID
 * OF THE SOFTWARE.
*/

#ifndef IFS_SCANNER_H
#define IFS_SCANNER_H

#include <list>
#include <string>
#include <systemc.h>

#include "IFS_Script.h"

namespace IFS
{
  /// Functions and data structures for handling a script on the lexical level
  namespace Scanner
  {

    /// A lexical element of the script
    struct Token
    {
      /// Constructor
      Token(std::string::size_type  tokenBegin,
            std::string::size_type  tokenEnd,
            const std::string &     tokenContent);

      /// Convenience method: compare m_TokenContent with s for equality
      bool operator == (const std::string &s) const;
      /// Convenience method: compare m_TokenContent with s for inequality
      bool operator != (const std::string &s) const;

      /// position of the first token character within the line (first one is 0)
      std::string::size_type  m_TokenBegin;
      /// position of the last token character within the line (first one is 0)
      std::string::size_type  m_TokenEnd;
      /// the token itself
      std::string             m_TokenContent;
    };

    /// Scanned version of SourceLine
    struct ScannedLine
    {
      /// convert scanned line back to a string
      std::string unparse() const;

      /// extract the m_TokenContent from the tokens and return them as a list
      std::list<std::string> getStringList() const;

      /// line number (first one is 1)
      unsigned int  m_LineNumber;
      /// Source line split into tokens
      std::list<Token>   m_LineContent;
      /// the file from which the line originates (\#inc)
      std::string        m_FileName;
    };


    /// Split source line into tokens
    /** \param sourceLine String containing the source line
      */
    ScannedLine scanSourceLine(const SourceLine &sourceLine);


    /// Check if \c s is an \<Identifier> (see Grammar)
    bool isIdentifier(const std::string &s);

    /// Check if \c s is an \<Integer> (see Grammar)
    bool isInteger(const std::string &s);

    /// Get value of an \<Integer>
    /// \pre isInteger(s)
    int getInteger(const std::string &s);

    /// Check if \c s is an \<PositiveInteger> (see Grammar)
    bool isPositiveInteger(const std::string &s);

    /// Check if \c s is an \<NonNegativeInteger> (see Grammar)
    bool isNonNegativeInteger(const std::string &s);

    /// Get value of an \<Integer>
    /// \pre isPositiveInteger(s)
    unsigned int getPositiveInteger(const std::string &s);

    /// Check if \c s is a time unit
    bool isTimeUnit(const std::string &s);

    /// Convert s to a sc_time_unit
    /// \pre isTimeUnit(s)
    sc_time_unit getTimeUnit(const std::string &s);

  }
}

// -----------------------------------------------------------------------------

inline
bool
IFS::Scanner::isInteger(const std::string &s)
{
  std::string::value_type const* p = s.c_str();

  while (*p != '\0')
  {
    if ((*p < '0' || *p > '9') && *p != '-' && *p != '+')
    {
      return false;
    }
    ++p;
  }

  return true;
}

#endif

