/*
 *  (C) Robert Bosch GmbH 2008. All rights reserved, also regarding any
 *  disposal, exploitation, reproduction, editing, distribution, as
 *  well as in the event of application for industrial property rights.
 *
 *  Contact: AE/NE1-IFS <vAENE1.IFS@de.bosch.com>
 *
 * Disclaimer:
 * This Software is licensed under the following conditions. You may 
 * not use this file except in compliance with the following conditions:
 *
 * Robert Bosch GmbH hereby grants to licensee the non-exclusive and 
 * non-transferable right to use the software and accompanying 
 * documentation for the sole purpose of testing designs delivered by 
 * Robert Bosch GmbH to licensee.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND WITHOUT ANY WARRANTY,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE, TITLE AND 
 * NONINFRINGEMENT. IN NO EVENT SHALL ROBERT BOSCH GMBH BE LIABLE FOR
 * ANY DIRECT OR INDIRECT DAMAGES, INJURY OR OTHER LIABILITY, ARISING
 * OUT OR IN CONNECTION WITH THE SOFTWARE. ROBERT BOSCH GMBH SHALL
 * NOT BE LIABLE FOR THE ACCURACY OF THE DESIGNS TESTED WITH THE AID
 * OF THE SOFTWARE.
*/

#ifndef IFS_SMARTPTR_H
#define IFS_SMARTPTR_H

#include <scv.h>
#include <stdarg.h>
#include <map>
#include <vector>
#include <string>
#include <sstream>
#include "IFS_Message.h"
#include "IFS_Utils.h"
#include "IFS_OutputHandler.h"

using namespace IFS;

namespace IFS {
  /*
   * BE CAREFUL: do not use KEEP_ONLY and KEEP_ONLY_LIST for the same constraint:
   * The following type of code will cause a segmentation fault:
   *
   * scv_smart_ptr<sc_uint <11> > temp;
   * list<sc_uint<11> > keep_list;
   * keep_list.push_back(11); keep_list.push_back(12); keep_list.push_back(13);
   * temp->keep_only(keep_list);
   * temp->reset_distribution();
   * temp->keep_only(5,15);
   *
   */

template <class T>
class ifs_smart_ptr {
  public:

    /// \brief
    /// check the coverage status of a constraint value
    /// \param  coverage_vector = vector elements are coverage signals / PSL-Endpoints
    /// \return Returns true, if everything is covered
    bool coverage_status(const vector<int>& coverage_vector) const
    {
      vector<int>::const_iterator iter = coverage_vector.begin();
      const vector<int>::const_iterator end = coverage_vector.end();
      while ((iter != end) && (*iter != 0))
      {
        ++iter;
      }
      return (iter == end);
    }

    /// \brief
    /// check whether the parameter value is adjusted
    /// \param  value = constraint value
    /// \return Returns true, if a the parameter value is adjusted
    bool adjusted(T value) const
    {
      typename vector<T>::const_iterator iter = adjusted_values.begin();
      const typename vector<T>::const_iterator end  = adjusted_values.end();
      while ((iter != end) && (*iter != value))
      {
        ++iter;
      }
      return (iter != end);
    }

    /// \brief
    /// adjust constraint value, if everything is covered
    /// \param  value = if everything is covered, "constraint will be adjusted to this value"
    /// \param  count = number of following coverage signals
    /// \param  last_named_arg = indicate the start of a variable parameter list
    void adjust(T value,int count, bool last_named_arg,...)
    {
      vector<int> coverage_vector(count);
      bool tmp_value = false;

      coverage_vector.push_back(last_named_arg);
      va_list param_ptr;
      va_start(param_ptr,last_named_arg);
      for(int i = 1; i < count ; ++i)
      {
        coverage_vector.push_back(va_arg(param_ptr, int));
      }
      va_end(param_ptr);

      if(coverage_status(coverage_vector))
      {
        stringstream s1;
        s1 << "ALL COVERED for Constraint: " <<
              s_ptr->get_name() << " = " << value;
        g_IfsOut << IFS_Message(s1.str(), IFS_Message::Utils, "" ,
                                IFS_Message::Info);
        if(adjusted(value))
        {
          stringstream s1;
          s1 << "REMOVING: " << value + " from " << s_ptr->get_name() <<
                " distribution";
          g_IfsOut << IFS_Message(s1.str(), IFS_Message::Utils, "" ,
                                  IFS_Message::Info);
        }
        else
        {
          if (bag.distinct_size() > 0)
          {
            stringstream s1;
            s1 << "REMOVING: " << value << " from " <<
                  s_ptr->get_name() << " distribution";
            g_IfsOut << IFS_Message(s1.str(), IFS_Message::Utils, "" ,
                                    IFS_Message::Info);
            while(!tmp_value)
            {
              if(value == bag.peekRandom())
              {
                bag.mark(true);
                tmp_value = true;
                s_ptr->reset_distribution();
                s_ptr->set_mode(bag);
              }
            }
          }
          else if (probability_bag.distinct_size() > 0)
          {
            stringstream s1;
            s1 << "REMOVING: " << value << " from " <<
                  s_ptr->get_name() << " distribution";
            g_IfsOut << IFS_Message(s1.str(), IFS_Message::Utils, "",
                                    IFS_Message::Info);
            while(!tmp_value)
            {
              if(value == probability_bag.peekRandom())
              {
                probability_bag.mark(true);
                tmp_value = true;
                s_ptr->reset_distribution();
                s_ptr->set_mode(probability_bag);
              }
            }
          }
          else if (range_bag.distinct_size() > 0)
          {
            stringstream s1;
            s1 << "REMOVING: " << value << " from " << s_ptr->get_name() << " distribution";
            g_IfsOut << IFS_Message(s1.str(), IFS_Message::Utils, "",
                                    IFS_Message::Info);
          }
          else //bag.distinct_size() == 0
          {
            stringstream s1;
            s1 << "UPDATING KEEP_OUT_LIST of  " << s_ptr->get_name() << " with: " << value;
            g_IfsOut << IFS_Message(s1.str(), IFS_Message::Utils, "",
                                    IFS_Message::Info);
            this->update_keep_out_list(value);
          }

          adjusted_values.push_back(value);
        } //if(adjusted(value))
      } //if(coverage_status(coverage_vector))
      else
      {
        stringstream s1;
        s1 << "NOT ALL COVERED for Constraint: " << s_ptr->get_name() << " = " << value;
        g_IfsOut << IFS_Message(s1.str(), IFS_Message::Utils, "",
                                IFS_Message::Info);
      }
    } //void adjust

    /// \brief
    /// adjust constraint value with default_value, if everything is covered
    /// \param  count = number of following coverage signals
    /// \param  last_named_arg = indicate the start of a variable parameter list
    void adjust_default(int count, bool last_named_arg,...)
    {
      vector<int> coverage_vector(count);
      coverage_vector.push_back(last_named_arg);

      va_list param_ptr;
      va_start(param_ptr,last_named_arg);
      for(int i = 1; i < count ; i++)
      {
        coverage_vector.push_back(va_arg(param_ptr, int));
      }
      va_end(param_ptr);

      if(coverage_status(coverage_vector))
      {
        stringstream s1;
        s1 << "ALL COVERED for Constraint: " << s_ptr->get_name();
        g_IfsOut << IFS_Message(s1.str(), IFS_Message::Utils, "",
                      IFS_Message::Info);

        if(!adjusted(default_value))
        {
          stringstream s2;
          s2 << "ADJUST CONSTRAINT: " << s_ptr->get_name() << " : " << default_value;
          g_IfsOut << IFS_Message(s2.str(), IFS_Message::Utils, "",
                    IFS_Message::Info);
          s_ptr->write(default_value);
          s_ptr->disable_randomization();
          adjusted_values.push_back(default_value);
        }
      } //if(coverage_status(coverage_vector))
      else
      {
        stringstream s1;
        s1 << "NOT ALL COVERED for Constraint: " << s_ptr->get_name() << " = " << default_value;
        g_IfsOut << IFS_Message(s1.str(), IFS_Message::Utils, "",
                    IFS_Message::Info);
      }
    }


    /// \brief
    /// enable randomization for this smart pointer
    void enable_randomization()
    {
      s_ptr->enable_randomization();
    }


    /// \brief
    /// disable randomization for this smart pointer
    void disable_randomization()
    {
      s_ptr->disable_randomization();
    }


    /// \brief
    /// deliver smart pointer name
    /// \return Return the name of the smart pointer
    string get_name()
    {
      return(s_ptr->get_name());
    }


    /// \brief
    /// set smart pointer name
    /// \param name = name of the smart pointer
    void set_name(string name)
    {
      s_ptr->set_name(name.c_str());
    }


    /// \brief
    /// smart pointer is set to a specific value
    /// function removes all former distributions etc.
    /// \ param value = from now on the smart pointer uses this value
    void update_keep_only(T value)
    {
      s_ptr->reset_distribution();
      s_ptr->keep_only(value);
    }


    /// \brief
    /// smart pointer only generates values between lowerbound and upperbound
    void update_keep_only(T lowerbound, T upperbound)
    {
      s_ptr->keep_only(lowerbound,upperbound);
    }


    /// \brief
    /// parameter value is left out during randomization
    void update_keep_out(T value)
    {
      s_ptr->keep_out(value);
    }


    /// \brief
    /// given range is left out during randomization
    void update_keep_out(T lowerbound, T upperbound)
    {
      s_ptr->keep_out(lowerbound,upperbound);
    }


    /// \brief
    /// update a list of keep_only values.
    /// randomization can only pick list elements
    /// \param value = added to keep_only list
    void update_keep_only_list(T value)
    {
      keep_only_list.push_back(value);
      s_ptr->reset_distribution();
      s_ptr->keep_only(keep_only_list);
    }


    /// \brief
    /// update a list of keep_out values.
    /// randomization cannot pick list elements
    /// \param value = added to keep_out list
    void update_keep_out_list(T value)
    {
      keep_out_list.push_back(value);
      s_ptr->keep_out(keep_out_list);
    }


    /// \brief
    /// update the distribution bag of the smart pointer
    /// randomization picks one of the bag elements
    /// \param value = added to smart pointer distribution
    void update_distribution(T value)
    {
      s_ptr->reset_distribution();
      bag.push(value);
      s_ptr->set_mode(bag);
    }


    /// \brief
    /// update the distribution bag of the smart pointer
    /// randomization picks one of the bag elements dependent on the probability
    /// \param value       = added to smart pointer distribution
    /// \param probability = number of generated bag entries for this value
    void update_distribution(T value, int probability)
    {
      s_ptr->reset_distribution();
      probability_bag.push(value,probability);
      s_ptr->set_mode(probability_bag);
    }


    /// \brief
    /// update the distribution bag of the smart pointer
    /// randomization picks one of the range-bag elements dependent on the probability
    /// lowerbound and upperbound used to describe a range
    /// \param lowerbound
    /// \param upperbound
    /// \param probability = number of generated bag entries for this range
    void update_distribution(T lowerbound,T upperbound, int probability)
    {
      s_ptr->reset_distribution();
      range_bag.push(pair<T,T>(lowerbound,upperbound),probability);
      s_ptr->set_mode(range_bag);
    }


    /// \brief
    /// delete lists/bags and reset distribution of smart pointer
    void clear()
    {
      keep_out_list.clear();
      keep_only_list.clear();
      range_bag.clear();
      bag.clear();
      probability_bag.clear();
      s_ptr->reset_distribution();
    }


    /// \brief
    /// read smart pointer value
    /// \return Returns the actual value of the smart pointer
    T read()
    {
      T value = s_ptr->read();
      return(value);
    }


    /// \brief
    /// write smart pointer value
    /// \param data
    void write(T data)
    {
      s_ptr->write(data);
    }


    /// \brief
    /// function is used for example in IFS_UPDATE_CONSTRAINT macro
    /// \return returns an address to a smart pointer
    scv_smart_ptr<T>* get_ptr()
    {
      return (&s_ptr);
    }

    /// \brief
    /// next random value
    void next()
    {
      s_ptr->next();
    }


    /// \brief
    /// ifs_smart_ptr usable in SCV_CONSTRAINT(..)
    /// \return returns a scv_expression
    scv_expression operator() ()
    {
      return s_ptr();
    }


    /// \brief
    /// return pointer
    /// \return returns pointer
    scv_smart_ptr<T>* operator* ()
    {
      return (&s_ptr);
    }

    /// default value, if randomization is disabled
    T                   default_value;

  private:
    /// randomization variable
    scv_smart_ptr<T>    s_ptr;
    /// store keep_out values
    list<T>             keep_out_list;
    /// store keep_only values
    list<T>             keep_only_list;
    /// used to create a distribution of several values
    scv_bag<T>          bag;
    /// assign a specific probability to a value
    scv_bag<T>          probability_bag;
    /// assign a specific probability to a range
    scv_bag<pair<T,T> > range_bag;

    /// store adjusted values of former simulation
    vector<T>           adjusted_values;
  };

}

#endif /*IFS_SMARTPTR_H*/

