/*
 *  (C) Robert Bosch GmbH 2008. All rights reserved, also regarding any
 *  disposal, exploitation, reproduction, editing, distribution, as
 *  well as in the event of application for industrial property rights.
 *
 *  Contact: AE/NE1-IFS <vAENE1.IFS@de.bosch.com>
 *
 * Disclaimer:
 * This Software is licensed under the following conditions. You may 
 * not use this file except in compliance with the following conditions:
 *
 * Robert Bosch GmbH hereby grants to licensee the non-exclusive and 
 * non-transferable right to use the software and accompanying 
 * documentation for the sole purpose of testing designs delivered by 
 * Robert Bosch GmbH to licensee.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND WITHOUT ANY WARRANTY,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE, TITLE AND 
 * NONINFRINGEMENT. IN NO EVENT SHALL ROBERT BOSCH GMBH BE LIABLE FOR
 * ANY DIRECT OR INDIRECT DAMAGES, INJURY OR OTHER LIABILITY, ARISING
 * OUT OR IN CONNECTION WITH THE SOFTWARE. ROBERT BOSCH GMBH SHALL
 * NOT BE LIABLE FOR THE ACCURACY OF THE DESIGNS TESTED WITH THE AID
 * OF THE SOFTWARE.
*/

#ifndef IFS_UTILS_H
#define IFS_UTILS_H

#include "IFS_ErrorMessage.h"
#include <string>
#include <systemc.h>
#include <map>

namespace IFS
{

  /// Converts a string into an integer.
  /** This function also supports prefixing of the numeric value with b#, d# or
    * h#.
    */
  int StringToInt( const std::string& sourceString ) throw( IFS_ErrorMessage );

  /// Convert string to unsigned int
  unsigned int StringToUInt( const std::string& sourceString ) throw( IFS_ErrorMessage );

  /// Converts a string into a double.
  /** This function also supports prefixing of the numeric value with b#, d# or
    * h#.
    */
  double StringToDouble( const std::string& sourceString ) throw( IFS_ErrorMessage );


  // Note: The following methods may also be realised as template'd version,
  //       although the result would have been returned via reference parameter.
  //       However, this would make the implementation visible in this header
  //       file and furthermore would force the inclusion of IFS_Expression.h.
  //       This may be problematic for the user due to multiple inclusion of
  //       systemc.h with and without SC_INCLUDE_FX.
  //       May be 'export' could help?

  /// Converts a string (ifs number) to a sc_int
  sc_int<64> StringToScInt64( const std::string& sourceString ) throw( IFS_ErrorMessage );

  /// Converts a string (ifs number) to a sc_uint
  sc_uint<64> StringToScUInt64( const std::string& sourceString ) throw( IFS_ErrorMessage );

  /// Converts a string (ifs number) to a sc_logic
  sc_logic StringToScLogic( const std::string& sourceString) throw( IFS_ErrorMessage );

  /// Converts a string (ifs number) to a sc_lv
  sc_lv<64> StringToScLv64( const std::string& sourceString) throw( IFS_ErrorMessage );

  /// Converts a string (ifs number) to a sc_bv
  sc_bv<64> StringToScBv64( const std::string& sourceString) throw( IFS_ErrorMessage );


  /// Converts a string (either "true" or "false") to a bool.
  bool StringToBool(const std::string& s) throw(IFS_ErrorMessage);

  /// Converts an int to a dec-string
  std::string IntToString(const int j);

  /// Converts an int to a bin-string
  std::string IntToBinString(const int j);

  /// Converts an int to a hex-string
  std::string IntToHexString(const int j);

  /// Converts an unsigned int to a dec-string
  std::string UIntToString(const unsigned int j);

  /// Converts an unsigned int to a bin-string
  std::string UIntToBinString(const unsigned int j);

  /// Converts an unsigned int to a hex-string
  std::string UIntToHexString(const unsigned int j);

  /// Converts a double to a dec-string
  std::string DoubleToString(const double j);

  /// Converts a sc_int<64> to a dec-string
  std::string ScInt64ToString(const sc_int<64>& j);

  /// Converts a sc_int<64> to a bin-string
  std::string ScInt64ToBinString(const sc_int<64>& j);

  /// Converts a sc_int<64> to a hex-string
  std::string ScInt64ToHexString(const sc_int<64>& j);

  /// Converts a sc_uint<64> to a dec-string
  std::string ScUInt64ToString(const sc_uint<64>& j);

  /// Converts a sc_uint<64> to a bin-string
  std::string ScUInt64ToBinString(const sc_uint<64>& j);

  /// Converts a sc_uint<64> to a hex-string
  std::string ScUInt64ToHexString(const sc_uint<64>& j);

  /// Converts a sc_lv to a dec-string
  std::string ScLvToString(const sc_lv_base& j);

  /// Converts a sc_lv to a bin-string
  std::string ScLvToBinString(const sc_lv_base& j);

  /// Converts a sc_lv to a hex-string
  std::string ScLvToHexString(const sc_lv_base& j);

  // Converts a sc_logic to a dec-string
  std::string ScLogicToString(const sc_logic& j);

  /// Converts a sc_logic to a bin-string
  std::string ScLogicToBinString(const sc_logic& j);

  /// Converts a sc_logic to a hex-string
  std::string ScLogicToHexString(const sc_logic& j);

  /// Converts a sc_bv to a dec-string
  std::string ScBvToString(const sc_bv_base& j);

  /// Converts a sc_bv to a bin-string
  std::string ScBvToBinString(const sc_bv_base& j);

  /// Converts a sc_bv to a hex-string
  std::string ScBvToHexString(const sc_bv_base& j);

  /// Stores a label and an address. The address has to be a positive IFS-Number without sign.
  void registerAddress(const std::string& label, const std::string& address) throw( IFS_ErrorMessage );

  /// For a given label, the corresponding address will be returned.
  std::string convLabelToAddress(const std::string& label) throw( IFS_ErrorMessage );

  /// For a given address, the corresponding label will be returned. The address has to be a positive IFS-Number without sign.
  std::string convAddressToLabel(const std::string& address) throw( IFS_ErrorMessage );
}

#endif

